from __future__ import annotations

import enum
from unittest import mock

import pytest

import zigpy.types as t
from zigpy.zcl.foundation import Status
import zigpy.zdo.types as zdo_t


@pytest.fixture
def expose_global():
    """`typing.get_type_hints` does not work for types defined within functions"""

    objects = []

    def inner(obj):
        assert obj.__name__ not in globals()
        globals()[obj.__name__] = obj

        objects.append(obj)

        return obj

    yield inner

    for obj in objects:
        del globals()[obj.__name__]


def test_enum_fields():
    class EnumNamed(t.enum8):
        NAME1 = 0x01
        NAME2 = 0x10

    assert EnumNamed("0x01") == EnumNamed.NAME1
    assert EnumNamed("1") == EnumNamed.NAME1
    assert EnumNamed("0x10") == EnumNamed.NAME2
    assert EnumNamed("16") == EnumNamed.NAME2
    assert EnumNamed("NAME1") == EnumNamed.NAME1
    assert EnumNamed("NAME2") == EnumNamed.NAME2
    assert EnumNamed("EnumNamed.NAME1") == EnumNamed.NAME1
    assert EnumNamed("EnumNamed.NAME2") == EnumNamed.NAME2


def test_struct_fields():
    class TestStruct(t.Struct):
        a: t.uint8_t
        b: t.uint16_t

    assert TestStruct.fields.a.name == "a"
    assert TestStruct.fields.a.type == t.uint8_t

    assert TestStruct.fields.b.name == "b"
    assert TestStruct.fields.b.type == t.uint16_t


def test_struct_subclass_creation():
    # In-class constants are allowed
    class TestStruct3(t.Struct):
        CONSTANT1: int = 123
        CONSTANT2 = 1234
        _private1: int = 456
        _private2 = 4567
        _PRIVATE_CONST = mock.sentinel.priv_const

        class Test:
            pass

    assert not TestStruct3.fields
    assert TestStruct3.CONSTANT1 == 123
    assert TestStruct3.CONSTANT2 == 1234
    assert TestStruct3._private1 == 456
    assert TestStruct3._private2 == 4567
    assert TestStruct3._PRIVATE_CONST is mock.sentinel.priv_const
    assert TestStruct3()._PRIVATE_CONST is mock.sentinel.priv_const
    assert TestStruct3.Test  # type: ignore[truthy-function]
    assert TestStruct3().Test
    assert "Test" not in TestStruct3().as_dict()

    # Still valid
    class TestStruct4(t.Struct):
        pass

    # Annotations with values are not fields
    class TestStruct5(t.Struct):
        a: t.uint8_t = 2  # not a field
        b: t.uint16_t  # is a field

    inst6 = TestStruct5(123)
    assert "a" not in inst6.as_dict()
    assert "b" in inst6.as_dict()

    # unless they are a StructField
    class TestStruct6(t.Struct):
        a: t.uint8_t = t.StructField()

    assert "a" in TestStruct6(2).as_dict()


def test_struct_construction():
    class TestStruct(t.Struct):
        a: t.uint8_t
        b: t.LVBytes

    s1 = TestStruct(a=1)
    s1.b = b"foo"

    s2 = TestStruct(a=1, b=b"foo")

    assert s1 == s2
    assert s1.a == s2.a
    assert s1.replace(b=b"foo") == s2.replace(b=b"foo")
    assert s1.serialize() == s2.serialize() == b"\x01\x03foo"

    assert TestStruct(s1) == s1

    # You cannot use the copy constructor with other keyword arguments
    with pytest.raises(ValueError):
        TestStruct(s1, b=b"foo")

    # Types are coerced on construction so you cannot pass bad values
    with pytest.raises(ValueError):
        TestStruct(a=object())

    # You can still assign bad values but serialization will fail
    s1.serialize()
    s1.b = object()

    with pytest.raises(ValueError):
        s1.serialize()


def test_nested_structs(expose_global):
    class OuterStruct(t.Struct):
        class InnerStruct(t.Struct):
            b: t.uint8_t
            c: t.uint8_t

        a: t.uint8_t
        inner: None = t.StructField(type=InnerStruct)
        d: t.uint8_t

    assert len(OuterStruct.fields) == 3
    assert OuterStruct.fields.a.type is t.uint8_t
    assert OuterStruct.fields.inner.type is OuterStruct.InnerStruct
    assert len(OuterStruct.fields.inner.type.fields) == 2
    assert OuterStruct.fields.d.type is t.uint8_t

    s, remaining = OuterStruct.deserialize(b"\x00\x01\x02\x03" + b"asd")
    assert remaining == b"asd"
    assert s.a == 0
    assert s.inner.b == 1
    assert s.inner.c == 2
    assert s.d == 3


def test_nested_structs2(expose_global):
    class OuterStruct(t.Struct):
        class InnerStruct(t.Struct):
            b: t.uint8_t
            c: t.uint8_t

        a: t.uint8_t
        inner: None = t.StructField(type=InnerStruct)
        d: t.uint8_t

    assert len(OuterStruct.fields) == 3
    assert OuterStruct.fields[0].type is t.uint8_t
    assert OuterStruct.fields[1].type is OuterStruct.InnerStruct
    assert len(OuterStruct.fields[1].type.fields) == 2
    assert OuterStruct.fields[2].type is t.uint8_t

    s, remaining = OuterStruct.deserialize(b"\x00\x01\x02\x03" + b"asd")
    assert remaining == b"asd"
    assert s.a == 0
    assert s.inner.b == 1
    assert s.inner.c == 2
    assert s.d == 3


def test_struct_init():
    class TestStruct(t.Struct):
        a: t.uint8_t
        b: t.uint16_t
        c: t.CharacterString

    ts = TestStruct(a=1, b=0x0100, c="TestStruct")
    assert repr(ts)
    assert isinstance(ts.a, t.uint8_t)
    assert isinstance(ts.b, t.uint16_t)
    assert isinstance(ts.c, t.CharacterString)
    assert ts.a == 1
    assert ts.b == 0x100
    assert ts.c == "TestStruct"

    ts2, remaining = TestStruct.deserialize(b"\x01\x00\x01\x0aTestStruct")
    assert not remaining
    assert ts == ts2
    assert ts.serialize() == ts2.serialize()

    ts3 = ts2.replace(b=0x0100)
    assert ts3 == ts2
    assert ts3.serialize() == ts2.serialize()

    ts4 = ts2.replace(b=0x0101)
    assert ts4 != ts2
    assert ts4.serialize() != ts2.serialize()


def test_struct_string_is_none():
    class TestStruct(t.Struct):
        a: t.CharacterString

    # str(None) == "None", which is bad
    with pytest.raises(ValueError):
        TestStruct(a=None).serialize()


def test_struct_field_dependencies():
    class TestStruct(t.Struct):
        foo: t.uint8_t
        status: Status
        bar: t.uint8_t = t.StructField(requires=lambda s: s.status == Status.SUCCESS)
        baz: t.uint8_t

    # Status is FAILURE so bar is not defined
    TestStruct(foo=1, status=Status.FAILURE, baz=2)

    ts1, remaining = TestStruct.deserialize(
        b"\x01" + Status.SUCCESS.serialize() + b"\x02\x03"
    )
    assert not remaining
    assert ts1 == TestStruct(foo=1, status=Status.SUCCESS, bar=2, baz=3)

    ts2, remaining = TestStruct.deserialize(
        b"\x01" + Status.FAILURE.serialize() + b"\x02\x03"
    )
    assert remaining == b"\x03"
    assert ts2 == TestStruct(foo=1, status=Status.FAILURE, bar=None, baz=2)


def test_struct_field_invalid_dependencies():
    class TestStruct(t.Struct):
        status: t.uint8_t
        value: t.uint8_t = t.StructField(requires=lambda s: s.status == 0x00)

    # Value will be ignored during serialization even though it has been assigned
    ts1 = TestStruct(status=0x01, value=0x02)
    assert ts1.serialize() == b"\x01"
    assert len(ts1.assigned_fields()) == 1

    # Value wasn't provided but it is required
    ts2 = TestStruct(status=0x00, value=None)
    assert len(ts1.assigned_fields()) == 1

    with pytest.raises(ValueError):
        ts2.serialize()

    # Value is not optional but doesn't need to be passed due to dependencies
    ts3 = TestStruct(status=0x01)
    assert ts3.serialize() == b"\x01"
    assert len(ts3.assigned_fields()) == 1


def test_struct_multiple_requires(expose_global):
    @expose_global
    class StrictStatus(t.enum8):
        SUCCESS = 0x00
        FAILURE = 0x01

        # Missing members cause a parsing failure
        _missing_ = enum.Enum._missing_

    class TestStruct(t.Struct):
        foo: t.uint8_t

        status1: StrictStatus
        value1: t.uint8_t = t.StructField(
            requires=lambda s: s.status1 == StrictStatus.SUCCESS
        )

        status2: StrictStatus
        value2: t.uint8_t = t.StructField(
            requires=lambda s: s.status2 == StrictStatus.SUCCESS
        )

    # status1: success, status2: success
    ts0, remaining = TestStruct.deserialize(
        b"\x00"
        + StrictStatus.SUCCESS.serialize()
        + b"\x01"
        + StrictStatus.SUCCESS.serialize()
        + b"\x02"
    )
    assert not remaining
    assert ts0 == TestStruct(
        foo=0,
        status1=StrictStatus.SUCCESS,
        value1=1,
        status2=StrictStatus.SUCCESS,
        value2=2,
    )

    # status1: failure, status2: success
    ts1, remaining = TestStruct.deserialize(
        b"\x00"
        + StrictStatus.FAILURE.serialize()
        + StrictStatus.SUCCESS.serialize()
        + b"\x02"
    )
    assert not remaining
    assert ts1 == TestStruct(
        foo=0, status1=StrictStatus.FAILURE, status2=StrictStatus.SUCCESS, value2=2
    )

    # status1: success, status2: failure, trailing
    ts2, remaining = TestStruct.deserialize(
        b"\x00"
        + StrictStatus.SUCCESS.serialize()
        + b"\x01"
        + StrictStatus.FAILURE.serialize()
        + b"\x02"
    )
    assert remaining == b"\x02"
    assert ts2 == TestStruct(
        foo=0, status1=StrictStatus.SUCCESS, value1=1, status2=StrictStatus.FAILURE
    )

    # status1: failure, status2: failure
    ts3, remaining = TestStruct.deserialize(
        b"\x00" + StrictStatus.FAILURE.serialize() + StrictStatus.FAILURE.serialize()
    )
    assert not remaining
    assert ts3 == TestStruct(
        foo=0, status1=StrictStatus.FAILURE, status2=StrictStatus.FAILURE
    )

    with pytest.raises(ValueError):
        # status1: failure
        TestStruct.deserialize(b"\x00" + StrictStatus.FAILURE.serialize())

    with pytest.raises(ValueError):
        # status1: failure, invalid trailing
        TestStruct.deserialize(b"\x00" + StrictStatus.FAILURE.serialize() + b"\xff")


def test_struct_equality():
    class TestStruct1(t.Struct):
        foo: t.uint8_t

    class TestStruct2(t.Struct):
        foo: t.uint8_t

    assert TestStruct1() != TestStruct2()
    assert TestStruct1(foo=1) != TestStruct2(foo=1)

    assert TestStruct1() == TestStruct1()
    assert TestStruct1(foo=1) == TestStruct1(foo=1)


@pytest.mark.parametrize(
    "data",
    [
        b"\x00",
        b"\x00\x00",
        b"\x01",
        b"\x01\x00",
        b"\x01\x02\x03",
        b"",
        b"\x00\x00\x00\x00",
    ],
)
def test_struct_subclass_extension(data):
    class TestStruct(t.Struct):
        foo: t.uint8_t

    class TestStructSubclass(TestStruct):
        bar: t.uint8_t = t.StructField(requires=lambda s: s.foo == 0x01)

    class TestCombinedStruct(t.Struct):
        foo: t.uint8_t
        bar: t.uint8_t = t.StructField(requires=lambda s: s.foo == 0x01)

    assert len(TestStructSubclass.fields) == 2
    assert len(TestCombinedStruct.fields) == 2

    error1 = None
    error2 = None

    try:
        ts1, remaining1 = TestStructSubclass.deserialize(data)
    except Exception as e:  # noqa: BLE001
        error1 = e

    try:
        ts2, remaining2 = TestCombinedStruct.deserialize(data)
    except Exception as e:  # noqa: BLE001
        error2 = e

    assert (error1 and error2) or (not error1 and not error2)

    if error1 or error2:
        assert repr(error1) == repr(error2)
    else:
        assert ts1.as_dict() == ts2.as_dict()
        assert remaining1 == remaining2


def test_optional_struct_special_case():
    class TestStruct(t.Struct):
        foo: t.uint8_t

    OptionalTestStruct = t.Optional(TestStruct)

    assert OptionalTestStruct.deserialize(b"") == (None, b"")
    assert OptionalTestStruct.deserialize(b"\x00") == (
        OptionalTestStruct(foo=0x00),
        b"",
    )


def test_conflicting_types():
    class GoodStruct(t.Struct):
        foo: t.uint8_t = t.StructField(type=t.uint8_t)

    with pytest.raises(TypeError):

        class BadStruct(t.Struct):
            foo: t.uint8_t = t.StructField(type=t.uint16_t)


def test_uppercase_field():
    class Neighbor(t.Struct):
        """Neighbor Descriptor"""

        PanId: t.EUI64
        IEEEAddr: t.EUI64
        NWKAddr: t.NWK
        NeighborType: t.uint8_t
        PermitJoining: t.uint8_t
        Depth: t.uint8_t
        LQI: t.uint8_t  # this should not be a constant

    assert len(Neighbor.fields) == 7
    assert Neighbor.fields[6].name == "LQI"
    assert Neighbor.fields[6].type == t.uint8_t


def test_non_annotated_field():
    with pytest.raises(TypeError):

        class TestStruct1(t.Struct):
            field1: t.uint8_t

            # Python does not provide any simple way to get the order of both defined
            # class attributes and annotations. This is bad.
            field2 = t.StructField(type=t.uint16_t)
            field3: t.uint32_t

    class TestStruct2(t.Struct):
        field1: t.uint8_t
        field2: None = t.StructField(type=t.uint16_t)
        field3: t.uint32_t

    assert len(TestStruct2.fields) == 3
    assert TestStruct2.fields[0] == t.StructField(name="field1", type=t.uint8_t)
    assert TestStruct2.fields[1] == t.StructField(name="field2", type=t.uint16_t)
    assert TestStruct2.fields[2] == t.StructField(name="field3", type=t.uint32_t)


def test_allowed_non_fields():
    class Other:
        def bar(self):
            return "bar"

    def foo2_(_):
        return "foo2"

    class TestStruct(t.Struct):
        @property
        def prop(self):
            return "prop"

        @prop.setter
        def prop(self, value):
            return

        foo1 = lambda _: "foo1"  # noqa: E731
        foo2 = foo2_
        bar = Other.bar

        field: t.uint8_t
        CONSTANT1: t.uint8_t = "CONSTANT1"
        CONSTANT2 = "CONSTANT2"

    assert len(TestStruct.fields) == 1
    assert TestStruct.CONSTANT1 == "CONSTANT1"
    assert TestStruct.CONSTANT2 == "CONSTANT2"
    assert TestStruct().prop == "prop"
    assert TestStruct().foo1() == "foo1"
    assert TestStruct().foo2() == "foo2"
    assert TestStruct().bar() == "bar"

    instance = TestStruct()
    instance.prop = None
    assert instance.prop == "prop"


def test_as_dict_empty_fields():
    class TestStruct(t.Struct):
        foo: t.uint8_t
        bar: t.uint8_t = t.StructField(requires=lambda s: s.foo == 0x01)

    assert TestStruct(foo=1, bar=2).as_dict() == {"foo": 1, "bar": 2}
    assert TestStruct(foo=0, bar=2).as_dict() == {"foo": 0, "bar": 2}
    assert TestStruct(foo=0).as_dict() == {"foo": 0, "bar": None}

    # Same thing as above but assigned as attributes
    ts1 = TestStruct()
    ts1.foo = 1
    ts1.bar = 2
    assert ts1.as_dict() == {"foo": 1, "bar": 2}

    ts2 = TestStruct()
    ts2.foo = 0
    ts2.bar = 2
    assert ts2.as_dict() == {"foo": 0, "bar": 2}

    ts3 = TestStruct()
    ts3.foo = 0
    assert ts3.as_dict() == {"foo": 0, "bar": None}


def test_no_types():
    with pytest.raises(TypeError):

        class TestBadStruct(t.Struct):
            field: None = t.StructField()


def test_repr():
    class TestStruct(t.Struct):
        foo: t.uint8_t

    assert repr(TestStruct(foo=1)) == "TestStruct(foo=1)"
    assert repr(TestStruct(foo=None)) == "TestStruct()"

    # Invalid values still work
    ts = TestStruct()
    ts.foo = 1j
    assert repr(ts) == "TestStruct(foo=1j)"


def test_repr_properties():
    class TestStruct(t.Struct):
        foo: t.uint8_t
        bar: t.uint8_t

        @property
        def baz(self):
            if self.bar is None:
                return None

            return t.Bool((self.bar & 0xF0) >> 4)

    assert repr(TestStruct(foo=1)) == "TestStruct(foo=1)"
    assert (
        repr(TestStruct(foo=1, bar=16))
        == "TestStruct(foo=1, bar=16, *baz=<Bool.true: 1>)"
    )
    assert repr(TestStruct()) == "TestStruct()"


def test_bitstruct_simple():
    class BitStruct1(t.Struct):
        foo: t.uint4_t
        bar: t.uint4_t

    s = BitStruct1(foo=0b1100, bar=0b1010)
    assert s.serialize() == bytes([0b1010_1100])

    s2, remaining = BitStruct1.deserialize(b"\x01\x02")
    assert remaining == b"\x02"
    assert s2.foo == 0b0001
    assert s2.bar == 0b0000


def test_bitstruct_nesting(expose_global):
    @expose_global
    class InnerBitStruct(t.Struct):
        baz1: t.uint1_t
        baz2: t.uint3_t
        baz3: t.uint1_t
        baz4: t.uint3_t

    class OuterStruct(t.Struct):
        foo: t.LVBytes
        bar: InnerBitStruct
        asd: t.uint8_t

    inner = InnerBitStruct(baz1=0b1, baz2=0b010, baz3=0b0, baz4=0b111)
    assert inner.serialize() == bytes([0b111_0_010_1])
    assert InnerBitStruct.deserialize(inner.serialize() + b"asd") == (inner, b"asd")

    s = OuterStruct(foo=b"asd", bar=inner, asd=0xFF)
    assert s.serialize() == b"\x03asd" + bytes([0b111_0_010_1]) + b"\xff"

    s2, remaining = OuterStruct.deserialize(s.serialize() + b"test")
    assert remaining == b"test"
    assert s == s2


def test_bitstruct_misaligned():
    class TestStruct(t.Struct):
        foo: t.uint1_t
        bar: t.uint8_t  # Even though this field is byte-serializable, it is misaligned
        baz: t.uint7_t

    s = TestStruct(foo=0b1, bar=0b10101010, baz=0b1110111)
    assert s.serialize() == bytes([0b1110111_1, 0b0101010_1])

    s2, remaining = TestStruct.deserialize(s.serialize() + b"asd")
    assert s == s2

    with pytest.raises(ValueError):
        TestStruct.deserialize(b"\xff")


def test_non_byte_sized_struct():
    class TestStruct(t.Struct):
        foo: t.uint1_t
        bar: t.uint8_t

    s = TestStruct(foo=1, bar=2)

    with pytest.raises(ValueError):
        s.serialize()

    with pytest.raises(ValueError):
        TestStruct.deserialize(b"\x00\x00\x00\x00")


def test_non_aligned_struct_non_integer_types():
    class TestStruct(t.Struct):
        foo: t.uint1_t
        bar: t.data8
        foo: t.uint7_t

    s = TestStruct(foo=1, bar=[2])

    with pytest.raises(ValueError):
        s.serialize()

    with pytest.raises(ValueError):
        TestStruct.deserialize(b"\x00\x00\x00\x00")


def test_bitstruct_complex():
    data = (
        b"\x11\x00\xff\xee\xdd\xcc\xbb\xaa\x08\x07\x06"
        b"\x05\x04\x03\x02\x01\x00\x00\x24\x02\x00\x7c"
    )

    neighbor, rest = zdo_t.Neighbor.deserialize(data + b"asd")
    assert rest == b"asd"

    neighbor2 = zdo_t.Neighbor(
        extended_pan_id=t.ExtendedPanId.convert("aa:bb:cc:dd:ee:ff:00:11"),
        ieee=t.EUI64.convert("01:02:03:04:05:06:07:08"),
        nwk=0x0000,
        device_type=zdo_t.Neighbor.DeviceType.Coordinator,
        rx_on_when_idle=zdo_t.Neighbor.RxOnWhenIdle.On,
        relationship=zdo_t.Neighbor.RelationShip.Sibling,
        reserved1=0b0,
        permit_joining=zdo_t.Neighbor.PermitJoins.Unknown,
        reserved2=0b000000,
        depth=0,
        lqi=124,
    )

    assert neighbor == neighbor2
    assert neighbor2.serialize() == data


def test_int_struct():
    class NonIntegralStruct(t.Struct):
        foo: t.uint8_t

    with pytest.raises(TypeError):
        int(NonIntegralStruct(123))

    # Integer structs must inherit from IntStruct
    with pytest.raises(TypeError):

        class BadIntegralStruct(t.Struct, t.uint8_t):
            foo: t.uint8_t

    # Integer structs must inherit from an integer type
    with pytest.raises(TypeError):

        class BadIntegralStruct2(t.IntStruct):
            foo: t.uint8_t

    class IntegralStruct(t.IntStruct, t.uint32_t):
        foo: t.uint8_t
        bar: t.uint16_t
        baz: t.uint7_t
        asd: t.uint1_t

    class IntegralStruct2(IntegralStruct):
        pass

    assert (
        IntegralStruct(0b0_1110001_1100110011001100_10101010)
        == IntegralStruct(
            foo=0b10101010,
            bar=0b1100110011001100,
            baz=0b1110001,
            asd=0b0,
        )
        == 0b0_1110001_1100110011001100_10101010
    )

    assert (
        IntegralStruct2(0b0_1110001_1100110011001100_10101010)
        == IntegralStruct2(
            foo=0b10101010,
            bar=0b1100110011001100,
            baz=0b1110001,
            asd=0b0,
        )
        == 0b0_1110001_1100110011001100_10101010
    )

    with pytest.raises(ValueError):
        # One extra bit
        IntegralStruct(0b1_0_1110001_1100110011001100_10101010)

    assert issubclass(IntegralStruct, t.uint32_t)
    assert issubclass(IntegralStruct, int)

    assert isinstance(IntegralStruct(1909247146), t.uint32_t)
    assert isinstance(IntegralStruct(1909247146), int)
    assert IntegralStruct(1909247146) == IntegralStruct(IntegralStruct(1909247146))

    # We do not accept anything but kwargs
    with pytest.raises(TypeError):
        assert IntegralStruct(1909247146, bar=0, baz=0, asd=0)

    # Or multiple positional arguments
    with pytest.raises(TypeError):
        assert IntegralStruct(1909247146, 0)


def test_struct_optional():
    class TestStruct(t.Struct):
        foo: t.uint8_t
        bar: t.uint16_t
        baz: t.uint8_t = t.StructField(requires=lambda s: s.bar == 2, optional=True)

    s1 = TestStruct(foo=1, bar=2, baz=3)
    assert s1.serialize() == b"\x01\x02\x00\x03"
    assert TestStruct.deserialize(s1.serialize() + b"asd") == (s1, b"asd")
    assert s1.replace(baz=None).serialize() == b"\x01\x02\x00"
    assert s1.replace(bar=4).serialize() == b"\x01\x04\x00"
    assert TestStruct.deserialize(b"\x01\x03\x00\x04") == (
        TestStruct(foo=1, bar=3),
        b"\x04",
    )


def test_struct_field_repr():
    class TestStruct(t.Struct):
        foo: t.uint8_t = t.StructField(repr=lambda v: v + 1)
        bar: t.uint16_t = t.StructField(repr=lambda v: "bar")
        baz: t.CharacterString = t.StructField(repr=lambda v: "baz")

    s1 = TestStruct(foo=1, bar=2, baz="asd")

    assert repr(s1) == "TestStruct(foo=2, bar=bar, baz=baz)"


def test_skip_missing():
    class TestStruct(t.Struct):
        foo: t.uint8_t
        bar: t.uint16_t

    assert TestStruct(foo=1).as_dict() == {"foo": 1, "bar": None}
    assert TestStruct(foo=1).as_dict(skip_missing=True) == {"foo": 1}

    assert TestStruct(foo=1).as_tuple() == (1, None)
    assert TestStruct(foo=1).as_tuple(skip_missing=True) == (1,)


def test_from_dict(expose_global):
    @expose_global
    class InnerStruct(t.Struct):
        field1: t.uint8_t
        field2: t.CharacterString

    class TestStruct(t.Struct):
        foo: t.uint8_t
        bar: InnerStruct
        baz: t.CharacterString

    s = TestStruct(foo=1, bar=InnerStruct(field1=2, field2="field2"), baz="field3")

    assert s == TestStruct.from_dict(s.as_dict(recursive=True))


def test_matching(expose_global):
    @expose_global
    class InnerStruct(t.Struct):
        field1: t.uint8_t
        field2: t.CharacterString

    class TestStruct(t.Struct):
        foo: t.uint8_t
        bar: InnerStruct
        baz: t.CharacterString

    assert TestStruct().matches(TestStruct())
    assert not TestStruct().matches(InnerStruct())
    assert TestStruct(foo=1).matches(TestStruct(foo=1))
    assert not TestStruct(foo=1).matches(TestStruct(foo=2))
    assert TestStruct(foo=1).matches(TestStruct())

    s = TestStruct(foo=1, bar=InnerStruct(field1=2, field2="asd"), baz="foo")
    assert s.matches(s)
    assert s.matches(TestStruct())
    assert s.matches(TestStruct(bar=InnerStruct()))
    assert s.matches(TestStruct(bar=InnerStruct(field1=2, field2="asd")))
    assert not s.matches(TestStruct(bar=InnerStruct(field1=3)))


def test_int_comparison(expose_global):
    @expose_global
    class FirmwarePlatform(t.enum8):
        Conbee = 0x05
        Conbee_II = 0x07
        Conbee_III = 0x09

    class FirmwareVersion(t.IntStruct, t.uint32_t):
        reserved: t.uint8_t
        platform: FirmwarePlatform
        minor: t.uint8_t
        major: t.uint8_t

    fw_ver = FirmwareVersion(0x264F0900)
    assert fw_ver == FirmwareVersion(
        reserved=0, platform=FirmwarePlatform.Conbee_III, minor=79, major=38
    )
    assert fw_ver == 0x264F0900
    assert int(fw_ver) == 0x264F0900
    assert "0x264F0900" in str(fw_ver)

    assert int(fw_ver) <= fw_ver
    assert fw_ver <= int(fw_ver)

    assert int(fw_ver) - 1 < fw_ver
    assert fw_ver < int(fw_ver) + 1

    assert int(fw_ver) >= fw_ver
    assert fw_ver >= int(fw_ver)

    assert int(fw_ver) + 1 > fw_ver
    assert fw_ver > int(fw_ver) - 1

    assert (fw_ver & 0b0010101) == (int(fw_ver) & 0b0010101)
    assert (fw_ver | 0b0010101) == (int(fw_ver) | 0b0010101)
    assert (fw_ver >> 3) == (int(fw_ver) >> 3)
    assert (fw_ver << 3) == (int(fw_ver) << 3)
    assert bool(fw_ver & 0) is False
    assert bool(fw_ver & 0xFFFF) is True
    assert hash(fw_ver) == hash(int(fw_ver))


def test_int_comparison_non_int(expose_global):
    @expose_global
    class FirmwarePlatform(t.enum8):
        Conbee = 0x05
        Conbee_II = 0x07
        Conbee_III = 0x09

    # This isn't an integer
    class FirmwareVersion(t.Struct):
        reserved: t.uint8_t
        platform: FirmwarePlatform
        minor: t.uint8_t
        major: t.uint8_t

    fw_ver = FirmwareVersion(
        reserved=0, platform=FirmwarePlatform.Conbee_III, minor=79, major=38
    )

    with pytest.raises(TypeError):
        fw_ver < 0  # noqa: B015

    with pytest.raises(TypeError):
        fw_ver <= 0  # noqa: B015

    with pytest.raises(TypeError):
        fw_ver > 0  # noqa: B015

    with pytest.raises(TypeError):
        fw_ver >= 0  # noqa: B015


def test_frozen_struct():
    class OuterStruct(t.Struct):
        class InnerStruct(t.Struct):
            b: t.uint8_t
            c: t.uint8_t

        a: t.uint8_t
        inner: None = t.StructField(type=InnerStruct)
        d: t.uint8_t
        e: t.uint16_t

    struct = OuterStruct(a=1, inner=OuterStruct.InnerStruct(b=2, c=3), d=4)
    frozen = struct.freeze()

    assert "frozen" not in repr(struct)
    assert "frozen" in repr(frozen)

    with pytest.raises(TypeError, match="Unhashable type"):
        hash(struct)

    # Setting attributes has no effect
    assert frozen.a == 1
    assert frozen.inner.b == 2

    with pytest.raises(AttributeError):
        frozen.a = 2

    with pytest.raises(AttributeError):
        frozen.inner.b = 5

    assert frozen.a == 1
    assert frozen.inner.b == 2

    assert {frozen: 2}[frozen] == 2
    assert {frozen, frozen} == {frozen}
    assert frozen == frozen.replace(a=1)
    assert {frozen, frozen, frozen.replace(a=1), frozen.replace(a=2)} == {
        frozen,
        frozen.replace(a=2),
    }


def test_struct_field_length():
    """Test that StructField `length` reads exactly N items from a List field."""

    class TestStruct(t.Struct):
        count: t.uint8_t
        items: t.List[t.uint16_t] = t.StructField(length=lambda s: s.count)

    s = TestStruct(count=3, items=[0x0001, 0x0002, 0x0003])
    serialized = s.serialize()
    assert serialized == b"\x03\x01\x00\x02\x00\x03\x00"

    s2, remaining = TestStruct.deserialize(serialized + b"extra")
    assert s2.count == 3
    assert list(s2.items) == [0x0001, 0x0002, 0x0003]
    assert remaining == b"extra"


def test_struct_field_length_zero():
    """Test that length=0 produces an empty list and consumes no data."""

    class TestStruct(t.Struct):
        count: t.uint8_t
        items: t.List[t.uint16_t] = t.StructField(length=lambda s: s.count)

    s = TestStruct(count=0, items=[])
    serialized = s.serialize()
    assert serialized == b"\x00"

    s2, remaining = TestStruct.deserialize(serialized + b"\xff\xff")
    assert s2.count == 0
    assert list(s2.items) == []
    assert remaining == b"\xff\xff"


def test_struct_field_length_with_trailing_field():
    """Test that length-limited List doesn't consume data needed by later fields."""

    class TestStruct(t.Struct):
        count: t.uint8_t
        items: t.List[t.uint16_t] = t.StructField(length=lambda s: s.count)
        trailer: t.uint8_t

    s = TestStruct(count=2, items=[0x000A, 0x000B], trailer=0xFF)
    serialized = s.serialize()
    assert serialized == b"\x02\x0a\x00\x0b\x00\xff"

    s2, remaining = TestStruct.deserialize(serialized + b"tail")
    assert s2.count == 2
    assert list(s2.items) == [0x000A, 0x000B]
    assert s2.trailer == 0xFF
    assert remaining == b"tail"


def test_struct_field_length_with_bitfields():
    """Test that length works with a bitfield count field."""

    class TestStruct(t.Struct):
        count: t.uint4_t
        flags: t.uint4_t
        items: t.List[t.uint16_t] = t.StructField(length=lambda s: s.count)

    s = TestStruct(count=2, flags=0x0F, items=[0x1234, 0x5678])
    serialized = s.serialize()
    # Bitfields: count=2 (low nibble), flags=0xF (high nibble) => 0xF2
    assert serialized == b"\xf2\x34\x12\x78\x56"

    s2, remaining = TestStruct.deserialize(serialized + b"rest")
    assert s2.count == 2
    assert s2.flags == 0x0F
    assert list(s2.items) == [0x1234, 0x5678]
    assert remaining == b"rest"


def test_struct_field_length_computed():
    """Test that length lambda can compute a value from other fields."""

    class TestStruct(t.Struct):
        raw_count: t.uint8_t
        items: t.List[t.uint16_t] = t.StructField(length=lambda s: s.raw_count * 2)

    s = TestStruct(raw_count=2, items=[0x0001, 0x0002, 0x0003, 0x0004])
    serialized = s.serialize()

    s2, remaining = TestStruct.deserialize(serialized + b"tail")
    assert s2.raw_count == 2
    assert list(s2.items) == [0x0001, 0x0002, 0x0003, 0x0004]
    assert remaining == b"tail"

    with pytest.raises(ValueError):
        # Not enough items provided
        TestStruct(raw_count=2, items=[0x0001]).serialize()

    # Zero is a special case
    assert (
        TestStruct(raw_count=0).serialize()
        == TestStruct(raw_count=0, items=[]).serialize()
        == b"\x00"
    )

    # Zero length with incongruent items array is an error
    with pytest.raises(ValueError):
        TestStruct(raw_count=0, items=[0x0001]).serialize()

    # Deserialization will fail if the exact number of elements isn't available
    with pytest.raises(ValueError):
        TestStruct.deserialize(s.serialize()[:-1])


def test_struct_field_bad_type_derived_length() -> None:
    """Test that structs cannot be created with non-List fields with lengths."""

    with pytest.raises(TypeError):

        class TestStruct(t.Struct):
            count: t.uint8_t
            items: t.uint16_t = t.StructField(length=lambda s: s.count)


def test_struct_field_default(expose_global):
    """Test that StructField `default` provides a default value for omitted fields."""

    @expose_global
    class Op(t.enum8):
        Read = 0x0
        Write = 0x1

    class TestStruct(t.Struct):
        value: t.uint8_t
        op: Op = t.StructField(default=Op.Read)

    # Omit op, should get default
    s1 = TestStruct(value=0x42)
    assert s1.op == Op.Read
    assert s1.serialize() == b"\x42\x00"

    # Provide op explicitly
    s2 = TestStruct(value=0x42, op=Op.Write)
    assert s2.op == Op.Write
    assert s2.serialize() == b"\x42\x01"



"""Zigbee Green Power Domain"""

from __future__ import annotations

from typing import Final

import zigpy.types as t
from zigpy.types.struct import StructField
from zigpy.zcl import Cluster, foundation
from zigpy.zcl.foundation import (
    BaseAttributeDefs,
    BaseCommandDefs,
    ZCLAttributeDef,
    ZCLCommandDef,
)
import zigpy.zgp.types as zgptypes


class CommissioningNotificationOptions(t.Struct):
    application_id: zgptypes.ApplicationID
    rx_after_tx: t.uint1_t
    security_level: zgptypes.SecurityLevel
    security_key_type: zgptypes.SecurityKeyType
    security_failed: t.uint1_t
    bidirectional_cap: t.uint1_t
    proxy_info_present: t.uint1_t
    _reserved: t.uint6_t


# Figure 27
class CommissioningNotificationSchema(foundation.CommandSchema):
    options: CommissioningNotificationOptions
    gpd_id: zgptypes.DeviceID
    frame_counter: t.uint32_t
    command_id: t.uint8_t
    payload: t.LVBytes
    gpp_short_addr: t.uint16_t = StructField(
        requires=lambda s: s.proxy_info_present, optional=True
    )
    distance: t.uint8_t = StructField(
        requires=lambda s: s.proxy_info_present, optional=True
    )
    mic: t.uint32_t = StructField(requires=lambda s: s.security_failed, optional=True)


class ResponseOptions(t.Struct):
    application_id: zgptypes.ApplicationID
    _reserved: t.uint5_t


# Figure 45
class ResponseSchema(foundation.CommandSchema):
    options: ResponseOptions
    temp_master_short_addr: t.uint16_t
    temp_master_tx_channel: t.uint8_t
    gpd_id: zgptypes.DeviceID
    gpd_command_id: t.uint8_t
    gpd_command_payload: t.LVBytes


# Figure 26
class PairingSearchOptions(t.Struct):
    application_id: zgptypes.ApplicationID
    request_unicast_sink: t.uint1_t
    request_derived_groupcast_sink: t.uint1_t
    request_commissioned_groupcast_sink: t.uint1_t
    request_frame_counter: t.uint1_t
    request_security_key: t.uint1_t
    _reserved: t.uint8_t


# Figure 25
class PairingSearchSchema(foundation.CommandSchema):
    options: PairingSearchOptions
    gpd_id: zgptypes.DeviceID


# Figure 24
class NotificationOptions(t.Struct):
    application_id: zgptypes.ApplicationID
    also_unicast: t.uint1_t
    also_derived_group: t.uint1_t
    also_commissioned_group: t.uint1_t
    security_level: zgptypes.SecurityLevel
    security_key_type: zgptypes.SecurityKeyType
    appoint_temp_master: t.uint1_t
    tx_queue_full: t.uint1_t
    _reserved: t.uint3_t


# Figure 23
class NotificationSchema(foundation.CommandSchema):
    options: NotificationOptions
    gpd_id: zgptypes.DeviceID
    frame_counter: t.uint32_t
    command_id: t.uint8_t
    payload: t.LVBytes
    short_addr: t.uint16_t = StructField(
        requires=lambda s: s.options.appoint_temp_master
    )
    distance: t.uint8_t = StructField(requires=lambda s: s.options.appoint_temp_master)


# Figure 40, 41
class PairingOptions(t.Struct):
    application_id: zgptypes.ApplicationID
    add_sink: t.uint1_t
    remove_gpd: t.uint1_t
    communication_mode: zgptypes.CommunicationMode
    gpd_fixed: t.uint1_t
    gpd_mac_seq_num_cap: t.uint1_t
    security_level: zgptypes.SecurityLevel
    security_key_type: zgptypes.SecurityKeyType
    security_frame_counter_present: t.uint1_t
    security_key_present: t.uint1_t
    assigned_alias_present: t.uint1_t
    forwarding_radius_present: t.uint1_t
    _reserved: t.uint6_t


# Figure 38, 39
class PairingSchema(foundation.CommandSchema):
    options: PairingOptions
    gpd_id: zgptypes.DeviceID
    # Table 37
    sink_ieee: t.EUI64 = StructField(
        requires=lambda s: not s.options.remove_gpd
        and s.options.communication_mode
        in (
            zgptypes.CommunicationMode.Unicast,
            zgptypes.CommunicationMode.UnicastLightweight,
        )
    )
    sink_nwk_addr: t.NWK = StructField(
        requires=lambda s: not s.options.remove_gpd
        and s.options.communication_mode
        in (
            zgptypes.CommunicationMode.Unicast,
            zgptypes.CommunicationMode.UnicastLightweight,
        )
    )
    sink_group: t.Group = StructField(
        requires=lambda s: not s.options.remove_gpd
        and s.options.communication_mode
        in (
            zgptypes.CommunicationMode.GroupcastForwardToDGroup,
            zgptypes.CommunicationMode.GroupcastForwardToCommGroup,
        )
    )

    device_id: t.uint8_t = StructField(requires=lambda s: s.options.add_sink)
    frame_counter: t.uint32_t = StructField(
        requires=lambda s: s.options.add_sink
        and s.options.security_frame_counter_present
    )
    key: t.KeyData = StructField(
        requires=lambda s: s.options.add_sink and s.options.security_key_present
    )
    alias: t.uint16_t = StructField(
        requires=lambda s: s.options.add_sink and s.options.assigned_alias_present
    )
    forwarding_radius: t.uint8_t = StructField(
        requires=lambda s: s.options.add_sink and s.options.forwarding_radius_present
    )


# Figure 37
class NotificationResponseOptions(t.Struct):
    application_id: zgptypes.ApplicationID
    first_to_forward: t.uint1_t
    no_pairing: t.uint1_t
    _reserved: t.uint3_t


class NotificationResponseSchema(foundation.CommandSchema):
    options: NotificationResponseOptions
    gpd_id: zgptypes.DeviceID
    frame_counter: t.uint32_t


# Figure 43
class ProxyCommissioningModeOptions(t.Struct):
    enter: t.uint1_t
    exit_mode: zgptypes.ProxyCommissioningModeExitMode
    channel_present: t.uint1_t
    unicast: t.uint1_t
    _reserved: t.uint2_t


class ProxyCommissioningModeSchema(foundation.CommandSchema):
    options: ProxyCommissioningModeOptions
    window: t.uint16_t = StructField(optional=True)


class GreenPowerProxy(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0021
    name: Final = "Green Power"
    ep_attribute: Final = "green_power"

    NotificationSchema: Final = NotificationSchema
    PairingSearchSchema: Final = PairingSearchSchema
    PairingSchema: Final = PairingSchema
    ResponseSchema: Final = ResponseSchema
    CommissioningNotificationSchema: Final = CommissioningNotificationSchema
    NotificationResponseSchema: Final = NotificationResponseSchema
    ProxyCommissioningModeSchema: Final = ProxyCommissioningModeSchema

    class AttributeDefs(BaseAttributeDefs):
        max_sink_table_entries: Final = ZCLAttributeDef(
            id=0x0000,
            type=t.uint8_t,
            access="r",
            mandatory=True,
        )
        sink_table: Final = ZCLAttributeDef(
            id=0x0001,
            type=t.LongOctetString,
            access="r",
            mandatory=True,
        )
        communication_mode: Final = ZCLAttributeDef(
            id=0x0002,
            type=zgptypes.CommunicationMode,
            access="rw",
            mandatory=True,
        )
        commissioning_exit_mode: Final = ZCLAttributeDef(
            id=0x0003,
            type=zgptypes.ProxyCommissioningModeExitMode,
            access="rw",
            mandatory=True,
        )
        commissioning_window: Final = ZCLAttributeDef(
            id=0x0004,
            type=t.uint16_t,
            access="rw",
        )
        security_level: Final = ZCLAttributeDef(
            id=0x0005,
            type=zgptypes.SecurityLevel,
            access="rw",
            mandatory=True,
        )
        functionality: Final = ZCLAttributeDef(
            id=0x0006,
            type=t.bitmap24,
            access="rw",
            mandatory=True,
        )
        active_functionality: Final = ZCLAttributeDef(
            id=0x0007,
            type=t.bitmap24,
            access="r",
            mandatory=True,
        )
        gpp_max_table_entries: Final = ZCLAttributeDef(
            id=0x0010, type=t.uint8_t, access="r"
        )
        gpp_proxy_table: Final = ZCLAttributeDef(
            id=0x0011, type=t.LongOctetString, access="r"
        )
        gpp_functionality: Final = ZCLAttributeDef(
            id=0x0016, type=t.bitmap24, access="r"
        )
        gpp_active_functionality: Final = ZCLAttributeDef(
            id=0x0017, type=t.bitmap24, access="r"
        )
        link_key: Final = ZCLAttributeDef(
            id=0x0022,
            type=t.KeyData,
            access="r",
            mandatory=True,
        )

    class ServerCommandDefs(BaseCommandDefs):
        notification: Final = ZCLCommandDef(
            id=0x00,
            schema=NotificationSchema,
        )

        pairing_search: Final = ZCLCommandDef(
            id=0x01,
            schema=PairingSearchSchema,
        )

        commissioning_notification: Final = ZCLCommandDef(
            id=0x04,
            schema=CommissioningNotificationSchema,
        )

    class ClientCommandDefs(BaseCommandDefs):
        notification_response: Final = ZCLCommandDef(
            id=0x00,
            schema=NotificationResponseSchema,
        )

        pairing: Final = ZCLCommandDef(
            id=0x01,
            schema=PairingSchema,
        )

        proxy_commissioning_mode: Final = ZCLCommandDef(
            id=0x02,
            schema=ProxyCommissioningModeSchema,
        )

        response: Final = ZCLCommandDef(
            id=0x06,
            schema=ResponseSchema,
        )



from __future__ import annotations

import enum
import logging
from typing import TYPE_CHECKING, Any

from zigpy.const import APS_REPLY_TIMEOUT
import zigpy.exceptions
import zigpy.profiles
import zigpy.types as t
import zigpy.util
import zigpy.zcl
from zigpy.zcl.foundation import GENERAL_COMMANDS, GeneralCommand, Status as ZCLStatus
from zigpy.zdo.types import Status as ZDOStatus

if TYPE_CHECKING:
    from zigpy.device import Device

LOGGER = logging.getLogger(__name__)


class Status(enum.IntEnum):
    """The status of an Endpoint"""

    # No initialization is done
    NEW = 0
    # Endpoint information (device type, clusters, etc) init done
    ZDO_INIT = 1
    # Endpoint Inactive
    ENDPOINT_INACTIVE = 3


class Endpoint(zigpy.util.LocalLogMixin, zigpy.util.ListenableMixin):
    """An endpoint on a device on the network"""

    def __init__(self, device: Device, endpoint_id: int) -> None:
        self._device: Device = device
        self._endpoint_id: int = endpoint_id
        self._listeners: dict = {}

        self.status: Status = Status.NEW
        self.profile_id: int | None = None
        self.device_type: zigpy.profiles.zha.DeviceType | None = None
        self.in_clusters: dict = {}
        self.out_clusters: dict = {}
        self._cluster_attr: dict = {}

        self._member_of: dict = {}

        self._manufacturer: str | None = None
        self._model: str | None = None

    async def initialize(self) -> None:
        self.info("Discovering endpoint information")

        if self.profile_id is not None or self.status == Status.ENDPOINT_INACTIVE:
            self.info("Endpoint descriptor already queried")
        else:
            status, _, sd = await self._device.zdo.Simple_Desc_req(
                self._device.nwk, self._endpoint_id
            )

            if status == ZDOStatus.NOT_ACTIVE:
                # These endpoints are essentially junk but this lets the device join
                self.status = Status.ENDPOINT_INACTIVE
                return
            elif status != ZDOStatus.SUCCESS:
                raise zigpy.exceptions.InvalidResponse(
                    "Failed to retrieve service descriptor: %s", status
                )

            self.info("Discovered endpoint information: %s", sd)
            self.profile_id = sd.profile
            self.device_type = sd.device_type

            if self.profile_id == zigpy.profiles.zha.PROFILE_ID:
                self.device_type = zigpy.profiles.zha.DeviceType(self.device_type)
            elif self.profile_id == zigpy.profiles.zll.PROFILE_ID:
                self.device_type = zigpy.profiles.zll.DeviceType(self.device_type)

            for cluster in sd.input_clusters:
                self.add_input_cluster(cluster)

            for cluster in sd.output_clusters:
                self.add_output_cluster(cluster)

        self.status = Status.ZDO_INIT

    @property
    def clusters(self) -> list[zigpy.zcl.Cluster]:
        """Return all clusters on this endpoint."""
        return [*self.in_clusters.values(), *self.out_clusters.values()]

    def add_input_cluster(
        self, cluster_id: int, cluster: zigpy.zcl.Cluster | None = None
    ) -> zigpy.zcl.Cluster:
        """Adds an endpoint's input cluster

        (a server cluster supported by the device)
        """
        if cluster is None:
            if cluster_id in self.in_clusters:
                return self.in_clusters[cluster_id]

            cluster = zigpy.zcl.Cluster.from_id(self, cluster_id, is_server=True)

        self.in_clusters[cluster_id] = cluster

        if cluster.ep_attribute is not None:
            self._cluster_attr[cluster.ep_attribute] = cluster

        if self._device.application._dblistener is not None:
            self._device.application._dblistener.register_cluster_events(cluster)

        return cluster

    def add_output_cluster(
        self, cluster_id: int, cluster: zigpy.zcl.Cluster | None = None
    ) -> zigpy.zcl.Cluster:
        """Adds an endpoint's output cluster

        (a client cluster supported by the device)
        """
        if cluster is None:
            if cluster_id in self.out_clusters:
                return self.out_clusters[cluster_id]

            cluster = zigpy.zcl.Cluster.from_id(self, cluster_id, is_server=False)

        self.out_clusters[cluster_id] = cluster

        if self._device.application._dblistener is not None:
            self._device.application._dblistener.register_cluster_events(cluster)

        return cluster

    async def add_to_group(self, grp_id: int, name: str | None = None) -> ZCLStatus:
        try:
            res = await self.groups.add(grp_id, name)
        except AttributeError:
            self.debug("Cannot add 0x%04x group, no groups cluster", grp_id)
            return ZCLStatus.FAILURE

        if res[0] not in (ZCLStatus.SUCCESS, ZCLStatus.DUPLICATE_EXISTS):
            self.debug("Couldn't add to 0x%04x group: %s", grp_id, res[0])
            return res[0]

        group = self.device.application.groups.add_group(grp_id, name)
        group.add_member(self)
        return res[0]

    async def remove_from_group(self, grp_id: int) -> ZCLStatus:
        try:
            res = await self.groups.remove(grp_id)
        except AttributeError:
            self.debug("Cannot remove 0x%04x group, no groups cluster", grp_id)
            return ZCLStatus.FAILURE

        if res[0] not in (ZCLStatus.SUCCESS, ZCLStatus.NOT_FOUND):
            self.debug("Couldn't remove to 0x%04x group: %s", grp_id, res[0])
            return res[0]

        if grp_id in self.device.application.groups:
            self.device.application.groups[grp_id].remove_member(self)
        return res[0]

    async def group_membership_scan(self) -> None:
        """Sync up group membership."""
        try:
            res = await self.groups.get_membership(groups=[])
        except AttributeError:
            return
        except (TimeoutError, zigpy.exceptions.ZigbeeException):
            self.debug("Failed to sync-up group membership")
            return

        if isinstance(res, GENERAL_COMMANDS[GeneralCommand.Default_Response].schema):
            self.debug("Device does not support group commands: %s", res)
            return

        groups = set(res[1])
        self.device.application.groups.update_group_membership(self, groups)

    async def get_model_info(self) -> tuple[str | None, str | None]:
        if zigpy.zcl.clusters.general.Basic.cluster_id not in self.in_clusters:
            return None, None

        # Some devices can't handle multiple attributes in the same read request
        for names in (["manufacturer", "model"], ["manufacturer"], ["model"]):
            try:
                success, failure = await self.basic.read_attributes(
                    names, allow_cache=True
                )
            except TimeoutError:
                # Only swallow the `TimeoutError` on the double attribute read
                if len(names) == 2:
                    continue

                raise

            if "model" in success:
                self._model = success["model"]

            if "manufacturer" in success:
                self._manufacturer = success["manufacturer"]

        return self._model, self._manufacturer

    async def request(
        self,
        cluster: t.ClusterId,
        sequence: t.uint8_t,
        data: bytes,
        command_id: GeneralCommand | t.uint8_t = 0x00,
        timeout=APS_REPLY_TIMEOUT,
        expect_reply: bool = True,
        use_ieee: bool = False,
        ask_for_ack: bool | None = None,
        priority: int | None = None,
        retries: int | None = None,
        retry_delay: float | None = None,
    ):
        if self.profile_id == zigpy.profiles.zll.PROFILE_ID and not (
            cluster == zigpy.zcl.clusters.lightlink.LightLink.cluster_id
            and command_id < 0x40
        ):
            profile_id = zigpy.profiles.zha.PROFILE_ID
        else:
            profile_id = self.profile_id

        return await self.device.request(
            profile=profile_id,
            cluster=cluster,
            src_ep=self._endpoint_id,
            dst_ep=self._endpoint_id,
            sequence=sequence,
            data=data,
            timeout=timeout,
            expect_reply=expect_reply,
            use_ieee=use_ieee,
            ask_for_ack=ask_for_ack,
            priority=priority,
            retries=retries,
            retry_delay=retry_delay,
        )

    async def reply(
        self,
        cluster: t.ClusterId,
        sequence: t.uint8_t,
        data: bytes,
        command_id: GeneralCommand | t.uint8_t = 0x00,
        timeout=APS_REPLY_TIMEOUT,
        expect_reply: bool = False,
        use_ieee: bool = False,
        ask_for_ack: bool | None = None,
        priority: int | None = None,
        retries: int | None = None,
        retry_delay: float | None = None,
    ) -> None:
        if self.profile_id == zigpy.profiles.zll.PROFILE_ID and not (
            cluster == zigpy.zcl.clusters.lightlink.LightLink.cluster_id
            and command_id < 0x40
        ):
            profile_id = zigpy.profiles.zha.PROFILE_ID
        else:
            profile_id = self.profile_id

        return await self.device.reply(
            profile=profile_id,
            cluster=cluster,
            src_ep=self._endpoint_id,
            dst_ep=self._endpoint_id,
            sequence=sequence,
            data=data,
            timeout=timeout,
            expect_reply=expect_reply,
            use_ieee=use_ieee,
            ask_for_ack=ask_for_ack,
            priority=priority,
            retries=retries,
            retry_delay=retry_delay,
        )

    def log(self, lvl: int, msg: str, *args: Any, **kwargs: Any) -> None:
        msg = "[0x%04x:%s] " + msg
        args = (self._device.nwk, self._endpoint_id, *args)
        LOGGER.log(lvl, msg, *args, **kwargs)

    @property
    def device(self) -> Device:
        return self._device

    @property
    def endpoint_id(self) -> int:
        return self._endpoint_id

    @property
    def manufacturer(self) -> str | None:
        if self._manufacturer is not None:
            return self._manufacturer
        return self.device.manufacturer

    @manufacturer.setter
    def manufacturer(self, value) -> None:
        self.warning(
            "Overriding manufacturer from quirks is not supported and "
            "will be removed in the next zigpy version"
        )
        self._manufacturer = value

    @property
    def manufacturer_id(self) -> int | None:
        """Return device's manufacturer id code."""
        return self.device.manufacturer_id

    @property
    def member_of(self) -> dict:
        return self._member_of

    @property
    def model(self) -> str | None:
        if self._model is not None:
            return self._model
        return self.device.model

    @model.setter
    def model(self, value) -> None:
        self.warning(
            "Overriding model from quirks is not supported and "
            "will be removed in the next version"
        )
        self._model = value

    @property
    def unique_id(self) -> tuple[t.EUI64, int]:
        return self.device.ieee, self.endpoint_id

    def __getattr__(self, name: str) -> zigpy.zcl.Cluster:
        try:
            return self._cluster_attr[name]
        except KeyError as exc:
            raise AttributeError from exc

    def __repr__(self) -> str:
        def cluster_repr(clusters):
            return ", ".join(
                [f"{c.ep_attribute}:0x{c.cluster_id:04X}" for c in clusters]
            )

        return (
            f"<{type(self).__name__}"
            f" id={self.endpoint_id}"
            f" in=[{cluster_repr(self.in_clusters.values())}]"
            f" out=[{cluster_repr(self.out_clusters.values())}]"
            f" status={self.status!r}"
            f">"
        )



import hashlib
from unittest.mock import AsyncMock, patch

import pytest

from tests.conftest import make_app
from zigpy.ota import OtaImageWithMetadata
import zigpy.ota.image
from zigpy.ota.providers import BaseOtaImageMetadata
from zigpy.zcl.clusters.general import Ota


@pytest.fixture
def image_with_metadata() -> OtaImageWithMetadata:
    firmware = zigpy.ota.image.OTAImage(
        header=zigpy.ota.image.OTAImageHeader(
            upgrade_file_id=zigpy.ota.image.OTAImageHeader.MAGIC_VALUE,
            file_version=0x12345678,
            image_type=0x5678,
            manufacturer_id=0x1234,
            header_version=256,
            header_length=60,
            field_control=zigpy.ota.image.FieldControl.HARDWARE_VERSIONS_PRESENT,
            minimum_hardware_version=1,
            maximum_hardware_version=5,
            stack_version=2,
            header_string="This is a test header!",
            image_size=60 + 2 + 4 + 8,
        ),
        subelements=[zigpy.ota.image.SubElement(tag_id=0x0000, data=b"fw_image")],
    )

    metadata = BaseOtaImageMetadata(
        file_version=0x12345678,
        manufacturer_id=0x1234,
        image_type=0x5678,
        checksum="sha256:" + hashlib.sha256(firmware.serialize()).hexdigest(),
        file_size=len(firmware.serialize()),
        manufacturer_names=("manufacturer1", "manufacturer2"),
        model_names=("model1", "model2"),
        changelog="Some simple changelog",
        min_hardware_version=1,
        max_hardware_version=5,
        min_current_file_version=0x12345678 - 10,
        max_current_file_version=0x12345678 - 2,
        specificity=0,
    )

    return OtaImageWithMetadata(metadata=metadata, firmware=firmware)


def test_ota_mirrored_metadata(image_with_metadata: OtaImageWithMetadata) -> None:
    assert image_with_metadata._min_hardware_version == 1
    assert image_with_metadata._max_hardware_version == 5
    assert image_with_metadata._manufacturer_id == 0x1234
    assert image_with_metadata._image_type == 0x5678

    # Metadata info is preferred so the firmware file itself isn't necessary
    image_with_no_firmware = image_with_metadata.replace(firmware=None)
    assert image_with_no_firmware._min_hardware_version == 1
    assert image_with_no_firmware._max_hardware_version == 5
    assert image_with_no_firmware._manufacturer_id == 0x1234
    assert image_with_no_firmware._image_type == 0x5678

    # But we can use it
    image_with_no_metadata_hw_versions = image_with_metadata.replace(
        metadata=image_with_metadata.metadata.replace(
            min_hardware_version=None,
            max_hardware_version=None,
            manufacturer_id=None,
            image_type=None,
        )
    )
    assert image_with_no_metadata_hw_versions._min_hardware_version == 1
    assert image_with_no_metadata_hw_versions._max_hardware_version == 5
    assert image_with_no_metadata_hw_versions._manufacturer_id == 0x1234
    assert image_with_no_metadata_hw_versions._image_type == 0x5678

    # Only if all are missing will the properties be `None`
    image_with_no_hw_versions = image_with_metadata.replace(
        metadata=image_with_metadata.metadata.replace(
            min_hardware_version=None,
            max_hardware_version=None,
            manufacturer_id=None,
            image_type=None,
        ),
        firmware=None,
    )
    assert image_with_no_hw_versions._min_hardware_version is None
    assert image_with_no_hw_versions._max_hardware_version is None
    assert image_with_no_hw_versions._manufacturer_id is None
    assert image_with_no_hw_versions._image_type is None


def test_metadata_specificity(image_with_metadata: OtaImageWithMetadata) -> None:
    def replace_meta(**kwargs):
        return image_with_metadata.replace(
            metadata=image_with_metadata.metadata.replace(**kwargs)
        )

    # If we lose useful metadata, the specificity decreases
    assert (
        0
        < replace_meta(manufacturer_names=(), model_names=()).specificity
        < replace_meta(manufacturer_names=()).specificity
        < replace_meta(max_current_file_version=None).specificity
        < image_with_metadata.specificity
    )


async def test_metadata_compatibility(
    image_with_metadata: OtaImageWithMetadata,
    make_initialized_device,
) -> None:
    app = make_app({})
    await app.initialize()

    dev = make_initialized_device(app)
    dev.model = "model1"
    dev.manufacturer = "manufacturer1"

    assert image_with_metadata.version == 0x12345678

    query_cmd = Ota.ServerCommandDefs.query_next_image.schema(
        field_control=Ota.QueryNextImageCommand.FieldControl.HardwareVersion,
        manufacturer_code=0x1234,
        image_type=0x5678,
        current_file_version=0x12345678 - 5,
        hardware_version=3,
    )

    assert image_with_metadata.check_compatibility(dev, query_cmd)

    # The file version is ignored when checking compatibility
    assert image_with_metadata.check_compatibility(
        dev, query_cmd.replace(current_file_version=0x12345678)
    )

    # The min and max current file versions are respected
    assert image_with_metadata.check_version(0x12345678 - 10)
    assert image_with_metadata.check_version(0x12345678 - 2)
    assert not image_with_metadata.check_version(0x12345678 - 11)
    assert not image_with_metadata.check_version(0x12345678 - 1)
    assert not image_with_metadata.check_version(0x12345678)

    assert not image_with_metadata.check_compatibility(
        dev, query_cmd.replace(image_type=0xAAAA)
    )

    assert not image_with_metadata.check_compatibility(
        dev, query_cmd.replace(manufacturer_code=0xAAAA)
    )

    with patch.object(dev, attribute="_model", new="model3"):
        assert not image_with_metadata.check_compatibility(dev, query_cmd)

    with patch.object(dev, attribute="_manufacturer", new="manufacturer3"):
        assert not image_with_metadata.check_compatibility(dev, query_cmd)

    assert not image_with_metadata.check_compatibility(
        dev, query_cmd.replace(hardware_version=0)
    )

    assert not image_with_metadata.check_compatibility(
        dev, query_cmd.replace(hardware_version=100)
    )

    # The image is super well-specified: if anything is missing, it becomes incompatible
    assert not image_with_metadata.check_compatibility(
        dev,
        query_cmd.replace(
            field_control=Ota.QueryNextImageCommand.FieldControl(0),
            hardware_version=None,
        ),
    )

    await app.shutdown()


async def test_metadata_fetch(image_with_metadata: OtaImageWithMetadata) -> None:
    image_without_firmware = image_with_metadata.replace(firmware=None)
    assert image_with_metadata.firmware is not None

    # Pretend we download the image contents
    object.__setattr__(
        image_without_firmware.metadata,
        "_fetch",
        AsyncMock(return_value=image_with_metadata.firmware.serialize()),
    )

    # New image is identical
    new_img = await image_without_firmware.fetch()
    assert new_img == image_with_metadata


@pytest.mark.parametrize(
    ("kwargs", "expected_trusted", "specificity_boost"),
    [
        ({}, False, 0),
        ({"trusted": False}, False, 0),
        ({"trusted": True}, True, 10000),
    ],
)
def test_metadata_trusted_specificity(
    image_with_metadata: OtaImageWithMetadata,
    kwargs: dict,
    expected_trusted: bool,
    specificity_boost: int,
) -> None:
    """Test trusted field and its effect on specificity."""
    img = image_with_metadata.replace(
        metadata=image_with_metadata.metadata.replace(**kwargs)
    )

    assert img.metadata.trusted is expected_trusted
    assert img.specificity == image_with_metadata.specificity + specificity_boost


def test_repr_with_firmware(image_with_metadata: OtaImageWithMetadata) -> None:
    """Test that repr summarizes firmware size instead of dumping raw bytes."""
    r = repr(image_with_metadata)
    assert r.startswith("OtaImageWithMetadata(metadata=")
    assert image_with_metadata.firmware is not None
    expected_size = image_with_metadata.firmware.header.image_size
    assert f"firmware=<OTAImage: {expected_size} bytes>" in r
    # Ensure the raw firmware bytes are not present
    assert "fw_image" not in r


def test_repr_without_firmware(image_with_metadata: OtaImageWithMetadata) -> None:
    """Test that repr handles missing firmware gracefully."""
    img = image_with_metadata.replace(firmware=None)
    r = repr(img)
    assert r.startswith("OtaImageWithMetadata(metadata=")
    assert "firmware=None)" in r



from __future__ import annotations

from typing import Final

import zigpy.types as t
from zigpy.zcl import Cluster
from zigpy.zcl.foundation import BaseCommandDefs, ZCLCommandDef


class LogicalType(t.enum2):
    Coordinator = 0b00
    Router = 0b01
    EndDevice = 0b10


class ZigbeeInformation(t.Struct):
    logical_type: LogicalType
    rx_on_when_idle: t.uint1_t
    reserved: t.uint5_t


class ScanRequestInformation(t.Struct):
    # whether the device is factory new
    factory_new: t.uint1_t

    # whether the device is capable of assigning addresses
    address_assignment: t.uint1_t
    reserved1: t.uint2_t

    # indicate the device is capable of initiating a link (i.e., it supports the
    # touchlink commissioning cluster at the client side) or 0 otherwise (i.e., it does
    # not support the touchlink commissioning cluster at the client side).
    touchlink_initiator: t.uint1_t
    undefined: t.uint1_t
    reserved2: t.uint1_t

    # If the ZLL profile is implemented, this bit shall be set to 0. In all other case
    # (Profile Interop / Zigbee 3.0), this bit shall be set to 1
    profile_interop: t.uint1_t


class ScanResponseInformation(t.Struct):
    factory_new: t.uint1_t
    address_assignment: t.uint1_t
    reserved1: t.uint2_t
    touchlink_initiator: t.uint1_t
    touchlink_priority_request: t.uint1_t
    reserved2: t.uint1_t
    profile_interop: t.uint1_t


class DeviceInfoRecord(t.Struct):
    ieee: t.EUI64
    endpoint_id: t.uint8_t
    profile_id: t.uint8_t
    device_id: t.uint16_t
    version: t.uint8_t
    group_id_count: t.uint8_t
    sort: t.uint8_t


class Status(t.enum8):
    Success = 0x00
    Failure = 0x01


class GroupInfoRecord(t.Struct):
    group_id: t.Group
    group_type: t.uint8_t


class EndpointInfoRecord(t.Struct):
    nwk_addr: t.NWK
    endpoint_id: t.uint8_t
    profile_id: t.uint16_t
    device_id: t.uint16_t
    version: t.uint8_t


class LightLink(Cluster):
    cluster_id: Final[t.uint16_t] = 0x1000
    ep_attribute: Final = "lightlink"

    class ServerCommandDefs(BaseCommandDefs):
        scan: Final = ZCLCommandDef(
            id=0x00,
            schema={
                "inter_pan_transaction_id": t.uint32_t,
                "zigbee_information": ZigbeeInformation,
                "touchlink_information": ScanRequestInformation,
            },
        )
        device_info: Final = ZCLCommandDef(
            id=0x02,
            schema={"inter_pan_transaction_id": t.uint32_t, "start_index": t.uint8_t},
        )
        identify: Final = ZCLCommandDef(
            id=0x06,
            schema={
                "inter_pan_transaction_id": t.uint32_t,
                "identify_duration": t.uint16_t,
            },
        )
        reset_to_factory_new: Final = ZCLCommandDef(
            id=0x07,
            schema={"inter_pan_transaction_id": t.uint32_t},
        )
        network_start: Final = ZCLCommandDef(
            id=0x10,
            schema={
                "inter_pan_transaction_id": t.uint32_t,
                "epid": t.EUI64,
                "key_index": t.uint8_t,
                "encrypted_network_key": t.KeyData,
                "logical_channel": t.uint8_t,
                "pan_id": t.PanId,
                "nwk_addr": t.NWK,
                "group_identifiers_begin": t.Group,
                "group_identifiers_end": t.Group,
                "free_network_addr_range_begin": t.NWK,
                "free_network_addr_range_end": t.NWK,
                "free_group_id_range_begin": t.Group,
                "free_group_id_range_end": t.Group,
                "initiator_ieee": t.EUI64,
                "initiator_nwk": t.NWK,
            },
        )
        network_join_router: Final = ZCLCommandDef(
            id=0x12,
            schema={
                "inter_pan_transaction_id": t.uint32_t,
                "epid": t.EUI64,
                "key_index": t.uint8_t,
                "encrypted_network_key": t.KeyData,
                "nwk_update_id": t.uint8_t,
                "logical_channel": t.uint8_t,
                "pan_id": t.PanId,
                "nwk_addr": t.NWK,
                "group_identifiers_begin": t.Group,
                "group_identifiers_end": t.Group,
                "free_network_addr_range_begin": t.NWK,
                "free_network_addr_range_end": t.NWK,
                "free_group_id_range_begin": t.Group,
                "free_group_id_range_end": t.Group,
            },
        )
        network_join_end_device: Final = ZCLCommandDef(
            id=0x14,
            schema={
                "inter_pan_transaction_id": t.uint32_t,
                "epid": t.EUI64,
                "key_index": t.uint8_t,
                "encrypted_network_key": t.KeyData,
                "nwk_update_id": t.uint8_t,
                "logical_channel": t.uint8_t,
                "pan_id": t.PanId,
                "nwk_addr": t.NWK,
                "group_identifiers_begin": t.Group,
                "group_identifiers_end": t.Group,
                "free_network_addr_range_begin": t.NWK,
                "free_network_addr_range_end": t.NWK,
                "free_group_id_range_begin": t.Group,
                "free_group_id_range_end": t.Group,
            },
        )
        network_update: Final = ZCLCommandDef(
            id=0x16,
            schema={
                "inter_pan_transaction_id": t.uint32_t,
                "epid": t.EUI64,
                "nwk_update_id": t.uint8_t,
                "logical_channel": t.uint8_t,
                "pan_id": t.PanId,
                "nwk_addr": t.NWK,
            },
        )
        # Utility
        get_group_identifiers: Final = ZCLCommandDef(
            id=0x41,
            schema={
                "start_index": t.uint8_t,
            },
        )
        get_endpoint_list: Final = ZCLCommandDef(
            id=0x42,
            schema={
                "start_index": t.uint8_t,
            },
        )

    class ClientCommandDefs(BaseCommandDefs):
        scan_rsp: Final = ZCLCommandDef(
            id=0x01,
            schema={
                "inter_pan_transaction_id": t.uint32_t,
                "rssi_correction": t.uint8_t,
                "zigbee_info": ZigbeeInformation,
                "touchlink_info": ScanResponseInformation,
                "key_bitmask": t.uint16_t,
                "response_id": t.uint32_t,
                "epid": t.EUI64,
                "nwk_update_id": t.uint8_t,
                "logical_channel": t.uint8_t,
                "pan_id": t.PanId,
                "nwk_addr": t.NWK,
                "num_sub_devices": t.uint8_t,
                "total_group_ids": t.uint8_t,
                "endpoint_id?": t.uint8_t,
                "profile_id?": t.uint16_t,
                "device_id?": t.uint16_t,
                "version?": t.uint8_t,
                "group_id_count?": t.uint8_t,
            },
        )
        device_info_rsp: Final = ZCLCommandDef(
            id=0x03,
            schema={
                "inter_pan_transaction_id": t.uint32_t,
                "num_sub_devices": t.uint8_t,
                "start_index": t.uint8_t,
                "device_info_records": t.LVList[DeviceInfoRecord, t.uint8_t],
            },
        )
        network_start_rsp: Final = ZCLCommandDef(
            id=0x11,
            schema={
                "inter_pan_transaction_id": t.uint32_t,
                "status": Status,
                "epid": t.EUI64,
                "nwk_update_id": t.uint8_t,
                "logical_channel": t.uint8_t,
                "pan_id": t.PanId,
            },
        )
        network_join_router_rsp: Final = ZCLCommandDef(
            id=0x13,
            schema={
                "inter_pan_transaction_id": t.uint32_t,
                "status": Status,
            },
        )
        network_join_end_device_rsp: Final = ZCLCommandDef(
            id=0x15,
            schema={
                "inter_pan_transaction_id": t.uint32_t,
                "status": Status,
            },
        )
        # Utility
        endpoint_info: Final = ZCLCommandDef(
            id=0x40,
            schema={
                "ieee_addr": t.EUI64,
                "nwk_addr": t.NWK,
                "endpoint_id": t.uint8_t,
                "profile_id": t.uint16_t,
                "device_id": t.uint16_t,
                "version": t.uint8_t,
            },
        )
        get_group_identifiers_rsp: Final = ZCLCommandDef(
            id=0x41,
            schema={
                "total": t.uint8_t,
                "start_index": t.uint8_t,
                "group_info_records": t.LVList[GroupInfoRecord, t.uint8_t],
            },
        )
        get_endpoint_list_rsp: Final = ZCLCommandDef(
            id=0x42,
            schema={
                "total": t.uint8_t,
                "start_index": t.uint8_t,
                "endpoint_info_records": t.LVList[EndpointInfoRecord, t.uint8_t],
            },
        )



from __future__ import annotations

import asyncio
import contextlib
from datetime import UTC, datetime, timedelta
import json
import logging
import re
import sqlite3
from typing import TYPE_CHECKING, Any, NamedTuple

import aiosqlite

import zigpy.appdb_schemas
import zigpy.backups
import zigpy.device
from zigpy.device import Device, Status as DeviceStatus
import zigpy.endpoint
from zigpy.endpoint import Endpoint, Status as EndpointStatus
import zigpy.exceptions
import zigpy.group
import zigpy.profiles
import zigpy.state
import zigpy.types as t
import zigpy.typing
from zigpy.typing import UNDEFINED
import zigpy.util
from zigpy.zcl import (
    AttributeClearedEvent,
    AttributeReadEvent,
    AttributeReportedEvent,
    AttributeUnsupportedEvent,
    AttributeUpdatedEvent,
    AttributeWrittenEvent,
    ClusterType,
    OtaQueryCacheClearedEvent,
    OtaQueryCacheUpdatedEvent,
)
from zigpy.zcl.clusters.general import Basic, Ota
from zigpy.zcl.foundation import Status
from zigpy.zdo import types as zdo_t

if TYPE_CHECKING:
    from zigpy.application import ControllerApplication
    from zigpy.zcl import Cluster


MIN_SQLITE_VERSION = (3, 24, 0)

if sqlite3.sqlite_version_info < MIN_SQLITE_VERSION:
    raise RuntimeError(
        f"zigpy requires SQLite {'.'.join(map(str, MIN_SQLITE_VERSION))} or newer "
        f"(found {sqlite3.sqlite_version})."
    )

LOGGER = logging.getLogger(__name__)

DB_VERSION = 15
DB_V = f"_v{DB_VERSION}"

UNIX_EPOCH = datetime.fromtimestamp(0, tz=UTC)
DB_V_REGEX = re.compile(r"(?:_v\d+)?$")

MIN_UPDATE_DELTA = timedelta(seconds=30).total_seconds()

# The old attribute cache was a simple `attrid: value` mapping. This works 99.9% of the
# time but unfortunately some devices reuse the same attribute ID on a standard cluster
# for two separate purposes, using a manufacturer code to distinguish them. We migrate
# attributes safely at runtime, once a device quirk has loaded and we can tell for sure
# if the device has "colliding" attributes.
UNMIGRATED_MANUFACTURER_CODE = -1


class AttributeCacheRow(NamedTuple):
    """A row from the attribute cache table. This format is internal and will change."""

    ieee: t.EUI64
    endpoint_id: int
    cluster_type: ClusterType
    cluster_id: int
    attr_id: int
    manufacturer_code: int | None
    status: Status
    value: Any
    last_updated: float


def _register_sqlite_adapters():
    def adapt_ieee(eui64):
        return str(eui64)

    sqlite3.register_adapter(t.EUI64, adapt_ieee)
    sqlite3.register_adapter(t.ExtendedPanId, adapt_ieee)

    def convert_ieee(s):
        return t.EUI64.convert(s.decode())

    sqlite3.register_converter("ieee", convert_ieee)


def aiosqlite_connect(
    database: str, iter_chunk_size: int = 64, **kwargs
) -> aiosqlite.Connection:
    """Copy of the the `aiosqlite.connect` function that connects using either the built-in
    `sqlite3` module or the imported `pysqlite3` module.
    """

    return aiosqlite.Connection(
        connector=lambda: sqlite3.connect(str(database), **kwargs),
        iter_chunk_size=iter_chunk_size,
    )


def decode_str_attribute(value: str | bytes) -> str:
    if isinstance(value, str):
        return value

    return value.split(b"\x00", 1)[0].decode("utf-8")


class PersistingListener(zigpy.util.CatchingTaskMixin):
    def __init__(
        self,
        connection: aiosqlite.Connection,
        application: ControllerApplication,
    ) -> None:
        _register_sqlite_adapters()

        self._db = connection
        self._application = application
        self._callback_handlers: asyncio.Queue = asyncio.Queue()
        self.running = False
        self._worker_task = asyncio.create_task(self._worker())

    async def initialize_tables(self) -> None:
        async with self.execute("PRAGMA integrity_check") as cursor:
            rows = await cursor.fetchall()
            status = "\n".join(row[0] for row in rows)

            if status != "ok":
                LOGGER.error(
                    "Zigbee database is corrupted, integrity check failed!\n%s", status
                )

        async with self.execute("PRAGMA foreign_key_check") as cursor:
            rows = await cursor.fetchall()

            if rows:
                LOGGER.error(
                    "Zigbee database is corrupted, foreign key check failed!\n%s", rows
                )

        # Truncate the SQLite journal file instead of deleting it after transactions
        await self._set_isolation_level(None)
        await self.execute("PRAGMA journal_mode = WAL")
        await self.execute("PRAGMA synchronous = normal")
        await self.execute("PRAGMA temp_store = memory")
        await self._set_isolation_level("DEFERRED")

        await self.execute("PRAGMA foreign_keys = ON")
        await self._run_migrations()

    @classmethod
    async def new(
        cls, database_file: str, app: ControllerApplication
    ) -> PersistingListener:
        """Create an instance of persisting listener."""
        sqlite_conn = await aiosqlite_connect(
            database_file,
            detect_types=sqlite3.PARSE_DECLTYPES,
            isolation_level="DEFERRED",  # The default is "", an alias for "DEFERRED"
        )
        listener = cls(sqlite_conn, app)

        try:
            await listener.initialize_tables()
        except Exception:  # noqa: BLE001
            await listener.shutdown()
            raise

        listener.running = True
        return listener

    async def _worker(self) -> None:
        """Process request in the received order."""
        while True:
            cb_name, args = await self._callback_handlers.get()
            handler = getattr(self, cb_name)
            assert handler
            try:
                await handler(*args)
            except sqlite3.Error as exc:
                LOGGER.debug(
                    "Error handling '%s' event with %s params: %s",
                    cb_name,
                    args,
                    str(exc),
                    exc_info=True,
                )
            except Exception:  # noqa: BLE001
                LOGGER.exception(
                    "Unexpected error while processing %s(%s)", cb_name, args
                )
            self._callback_handlers.task_done()

    async def shutdown(self) -> None:
        """Shutdown connection."""
        self.running = False
        await self._callback_handlers.join()
        if not self._worker_task.done():
            self._worker_task.cancel()

        # Delete the journal on shutdown
        await self._set_isolation_level(None)
        await self.execute("PRAGMA wal_checkpoint;")
        await self._set_isolation_level("DEFERRED")

        await self._db.close()

        # FIXME: aiosqlite's thread won't always be closed immediately
        await asyncio.get_running_loop().run_in_executor(None, self._db.join)

    def register_cluster_events(self, cluster) -> None:
        cluster.on_event(AttributeReadEvent.event_type, self.on_attribute_read)
        cluster.on_event(AttributeReportedEvent.event_type, self.on_attribute_reported)
        cluster.on_event(AttributeUpdatedEvent.event_type, self.on_attribute_updated)
        cluster.on_event(AttributeWrittenEvent.event_type, self.on_attribute_written)
        cluster.on_event(
            AttributeUnsupportedEvent.event_type, self.on_attribute_unsupported
        )
        cluster.on_event(AttributeClearedEvent.event_type, self.on_attribute_cleared)
        cluster.on_event(
            OtaQueryCacheUpdatedEvent.event_type,
            self.on_ota_query_cache_updated,
        )
        cluster.on_event(
            OtaQueryCacheClearedEvent.event_type,
            self.on_ota_query_cache_cleared,
        )

    def enqueue(self, cb_name: str, *args) -> None:
        """Enqueue an async callback handler action."""
        if not self.running:
            LOGGER.debug("Discarding %s event", cb_name)
            return
        self._callback_handlers.put_nowait((cb_name, args))

    async def _set_isolation_level(self, level: str | None):
        """Set the SQLite statement isolation level in a thread-safe way."""
        await self._db._execute(lambda: setattr(self._db, "isolation_level", level))

    def execute(self, *args, **kwargs):
        return self._db.execute(*args, **kwargs)

    async def executescript(self, sql):
        """Naive replacement for `sqlite3.Cursor.executescript` that does not execute a
        `COMMIT` before running the script. This extra `COMMIT` breaks transactions that
        run scripts.
        """

        # XXX: This will break if you use a semicolon anywhere but at the end of a line
        for statement in sql.split(";"):
            # Strip SQL comments so that pysqlite3's implicit transaction DML detection
            # (which doesn't skip comments) sees the actual statement keyword
            statement = re.sub(r"--[^\n]*", "", statement)

            await self.execute(statement)

    def device_joined(self, device: Device) -> None:
        self.enqueue("_update_device_nwk", device.ieee, device.nwk)

    async def _update_device_nwk(self, ieee: t.EUI64, nwk: t.NWK) -> None:
        await self.execute(f"UPDATE devices{DB_V} SET nwk=? WHERE ieee=?", (nwk, ieee))
        await self._db.commit()

    def device_initialized(self, device: Device) -> None:
        pass

    def device_left(self, device: Device) -> None:
        pass

    def device_last_seen_updated(self, device: Device, last_seen: datetime) -> None:
        """Device last_seen time is updated."""
        self.enqueue("_save_device_last_seen", device.ieee, last_seen)

    async def _save_device_last_seen(self, ieee: t.EUI64, last_seen: datetime) -> None:
        q = f"""UPDATE devices{DB_V}
                    SET last_seen=:ts
                    WHERE ieee=:ieee AND :ts - last_seen > :min_update_delta"""
        await self.execute(
            q,
            {
                "ts": last_seen.timestamp(),
                "ieee": ieee,
                "min_update_delta": MIN_UPDATE_DELTA,
            },
        )
        await self._db.commit()

    def device_relays_updated(self, device: Device, relays: t.Relays | None) -> None:
        """Device relay list is updated."""
        self.enqueue("_save_device_relays", device.ieee, relays)

    async def _save_device_relays(self, ieee: t.EUI64, relays: t.Relays | None) -> None:
        if relays is None:
            await self.execute(f"DELETE FROM relays{DB_V} WHERE ieee = ?", (ieee,))
        else:
            q = f"""INSERT INTO relays{DB_V} VALUES (:ieee, :relays)
                        ON CONFLICT (ieee)
                        DO UPDATE SET relays=excluded.relays WHERE relays != :relays"""
            await self.execute(q, {"ieee": ieee, "relays": relays.serialize()})

        await self._db.commit()

    def neighbors_updated(self, ieee: t.EUI64, neighbors: list[zdo_t.Neighbor]) -> None:
        """Neighbor update from Mgmt_Lqi_req."""
        self.enqueue("_neighbors_updated", ieee, neighbors)

    async def _neighbors_updated(
        self, ieee: t.EUI64, neighbors: list[zdo_t.Neighbor]
    ) -> None:
        await self.execute(f"DELETE FROM neighbors{DB_V} WHERE device_ieee = ?", [ieee])

        rows = [(ieee, *neighbor.as_tuple()) for neighbor in neighbors]

        await self._db.executemany(
            f"INSERT INTO neighbors{DB_V} VALUES (?,?,?,?,?,?,?,?,?,?,?,?)", rows
        )
        await self._db.commit()

    def routes_updated(self, ieee: t.EUI64, routes: list[zdo_t.Route]) -> None:
        """Route update from Mgmt_Rtg_req."""
        self.enqueue("_routes_updated", ieee, routes)

    async def _routes_updated(self, ieee: t.EUI64, routes: list[zdo_t.Route]) -> None:
        await self.execute(f"DELETE FROM routes{DB_V} WHERE device_ieee = ?", [ieee])

        rows = [(ieee, *route.as_tuple()) for route in routes]

        await self._db.executemany(
            f"INSERT INTO routes{DB_V} VALUES (?,?,?,?,?,?,?,?)", rows
        )
        await self._db.commit()

    def group_added(self, group: zigpy.group.Group) -> None:
        """Group is added."""
        self.enqueue("_group_added", group)

    async def _group_added(self, group: zigpy.group.Group) -> None:
        q = f"""INSERT INTO groups{DB_V} VALUES (?, ?)
                    ON CONFLICT (group_id)
                    DO UPDATE SET name=excluded.name"""
        await self.execute(q, (group.group_id, group.name))
        await self._db.commit()

    def group_member_added(self, group: zigpy.group.Group, ep: Endpoint) -> None:
        """Called when a group member is added."""
        self.enqueue("_group_member_added", group, ep)

    async def _group_member_added(self, group: zigpy.group.Group, ep: Endpoint) -> None:
        q = f"""INSERT INTO group_members{DB_V} VALUES (?, ?, ?)
                    ON CONFLICT
                    DO NOTHING"""
        await self.execute(q, (group.group_id, *ep.unique_id))
        await self._db.commit()

    def group_member_removed(self, group: zigpy.group.Group, ep: Endpoint) -> None:
        """Called when a group member is removed."""
        self.enqueue("_group_member_removed", group, ep)

    async def _group_member_removed(
        self, group: zigpy.group.Group, ep: Endpoint
    ) -> None:
        q = f"""DELETE FROM group_members{DB_V} WHERE group_id=?
                                                AND ieee=?
                                                AND endpoint_id=?"""
        await self.execute(q, (group.group_id, *ep.unique_id))
        await self._db.commit()

    def group_removed(self, group: zigpy.group.Group) -> None:
        """Called when a group is removed."""
        self.enqueue("_group_removed", group)

    async def _group_removed(self, group: zigpy.group.Group) -> None:
        q = f"DELETE FROM groups{DB_V} WHERE group_id=?"
        await self.execute(q, (group.group_id,))
        await self._db.commit()

    def device_removed(self, device: Device) -> None:
        self.enqueue("_remove_device", device)

    async def _remove_device(self, device: Device) -> None:
        await self.execute(f"DELETE FROM devices{DB_V} WHERE ieee = ?", (device.ieee,))
        await self._db.commit()

    def raw_device_initialized(self, device: Device) -> None:
        # We work with a clone of a device so that quirks that run in the same event
        # loop iteration do not manage to mutate it before the async database code runs
        self.enqueue("_raw_device_initialized", device.clone())

    async def _raw_device_initialized(self, device: Device) -> None:
        # To ensure events do not leak with the discarded device clone, we need to clean
        # up
        try:
            await self._raw_device_initialized_internal(device)
        finally:
            device.on_remove()

    async def _raw_device_initialized_internal(self, device: Device) -> None:
        # `device` is a clone snapshotted in `raw_device_initialized`, taken before any
        # quirk had a chance to mutate the live device
        if device.original_signature is not None:
            raise ValueError("A device with quirks cannot be saved to the database")

        q = f"""INSERT INTO devices{DB_V} (ieee, nwk, status, last_seen)
                    VALUES (?, ?, ?, ?)
                    ON CONFLICT (ieee)
                    DO UPDATE SET
                        nwk=excluded.nwk,
                        status=excluded.status,
                        last_seen=excluded.last_seen"""
        await self.execute(
            q,
            (
                device.ieee,
                device.nwk,
                device.status,
                (device._last_seen or UNIX_EPOCH).timestamp(),
            ),
        )

        if device.node_desc is not None:
            await self._save_node_descriptor(device)

        await self._save_endpoints(device)
        for ep in device.non_zdo_endpoints:
            await self._save_clusters(ep)
            await self._save_attribute_cache(ep)
            await self._save_unsupported_attributes(ep)
        await self._save_ota_query_cache(device)
        await self._db.commit()

    async def _save_endpoints(self, device: Device) -> None:
        rows = [
            (
                device.ieee,
                ep.endpoint_id,
                ep.profile_id,
                ep.device_type,
                ep.status,
            )
            for ep in device.non_zdo_endpoints
        ]

        q = f"""INSERT INTO endpoints{DB_V} VALUES (?, ?, ?, ?, ?)
                    ON CONFLICT (ieee, endpoint_id)
                    DO UPDATE SET
                        profile_id=excluded.profile_id,
                        device_type=excluded.device_type,
                        status=excluded.status"""

        await self._db.executemany(q, rows)

    async def _save_node_descriptor(self, device: Device) -> None:
        if device.node_desc is None:
            return

        q = f"""INSERT INTO node_descriptors{DB_V}
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                    ON CONFLICT (ieee)
                    DO UPDATE SET
                logical_type=excluded.logical_type,
                complex_descriptor_available=excluded.complex_descriptor_available,
                user_descriptor_available=excluded.user_descriptor_available,
                reserved=excluded.reserved,
                aps_flags=excluded.aps_flags,
                frequency_band=excluded.frequency_band,
                mac_capability_flags=excluded.mac_capability_flags,
                manufacturer_code=excluded.manufacturer_code,
                maximum_buffer_size=excluded.maximum_buffer_size,
                maximum_incoming_transfer_size=excluded.maximum_incoming_transfer_size,
                server_mask=excluded.server_mask,
                maximum_outgoing_transfer_size=excluded.maximum_outgoing_transfer_size,
                descriptor_capability_field=excluded.descriptor_capability_field"""

        await self.execute(q, (device.ieee, *device.node_desc.as_tuple()))

    async def _save_clusters(self, endpoint: Endpoint) -> None:
        clusters = [
            (
                endpoint.device.ieee,
                endpoint.endpoint_id,
                cluster.cluster_type,
                cluster.cluster_id,
            )
            for cluster in endpoint.clusters
        ]
        q = f"""INSERT INTO clusters{DB_V} VALUES (?, ?, ?, ?)
                    ON CONFLICT (ieee, endpoint_id, cluster_type, cluster_id)
                    DO NOTHING"""
        await self._db.executemany(q, clusters)

    async def _save_attribute_cache(self, ep: Endpoint) -> None:
        clusters = [
            (
                ep.device.ieee,
                ep.endpoint_id,
                cluster.cluster_type,
                cluster.cluster_id,
                attrid,
                manufacturer_code,
                Status.SUCCESS,
                cache_item.value,
                cache_item.last_updated.timestamp(),
            )
            for cluster in ep.clusters
            for (
                attrid,
                manufacturer_code,
            ), cache_item in cluster._attr_cache._cache.items()
        ]
        q = f"""INSERT INTO attributes_cache{DB_V} (ieee, endpoint_id, cluster_type, cluster_id, attr_id, manufacturer_code, status, value, last_updated)
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
                    ON CONFLICT (ieee, endpoint_id, cluster_type, cluster_id, attr_id, manufacturer_code_idx)
                    DO UPDATE SET status=excluded.status, value=excluded.value, last_updated=excluded.last_updated"""
        await self._db.executemany(q, clusters)

    async def _save_unsupported_attributes(self, ep: Endpoint) -> None:
        clusters = [
            (
                ep.device.ieee,
                ep.endpoint_id,
                cluster.cluster_type,
                cluster.cluster_id,
                attrid,
                manufacturer_code,
                Status.UNSUPPORTED_ATTRIBUTE,
                None,
                datetime.now(UTC).timestamp(),
            )
            for cluster in ep.clusters
            for (attrid, manufacturer_code) in cluster._attr_cache._unsupported
        ]
        q = f"""INSERT INTO attributes_cache{DB_V} (ieee, endpoint_id, cluster_type, cluster_id, attr_id, manufacturer_code, status, value, last_updated)
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
                    ON CONFLICT (ieee, endpoint_id, cluster_type, cluster_id, attr_id, manufacturer_code_idx)
                    DO NOTHING"""
        await self._db.executemany(q, clusters)

    async def _save_ota_query_cache(self, device: Device) -> None:
        rows = []
        for ep in device.non_zdo_endpoints:
            for cluster in ep.clusters:
                if isinstance(cluster, Ota) and cluster.last_query_cmd is not None:
                    cmd = cluster.last_query_cmd
                    rows.append(
                        (
                            device.ieee,
                            ep.endpoint_id,
                            cmd.manufacturer_code,
                            cmd.image_type,
                            cmd.current_file_version,
                            getattr(cmd, "hardware_version", None),
                            datetime.now(UTC).timestamp(),
                        )
                    )
        if rows:
            q = f"""INSERT INTO ota_query_cache{DB_V}
                        (ieee, endpoint_id, manufacturer_code, image_type,
                         current_file_version, hardware_version, last_updated)
                    VALUES (?, ?, ?, ?, ?, ?, ?)
                    ON CONFLICT (ieee, endpoint_id) DO UPDATE SET
                        manufacturer_code=excluded.manufacturer_code,
                        image_type=excluded.image_type,
                        current_file_version=excluded.current_file_version,
                        hardware_version=excluded.hardware_version,
                        last_updated=excluded.last_updated"""
            await self._db.executemany(q, rows)

    def on_attribute_read(self, event: AttributeReadEvent) -> None:
        self.enqueue("_save_attribute", event)

    def on_attribute_reported(self, event: AttributeReportedEvent) -> None:
        self.enqueue("_save_attribute", event)

    def on_attribute_updated(self, event: AttributeUpdatedEvent) -> None:
        self.enqueue("_save_attribute", event)

    def on_attribute_written(self, event: AttributeWrittenEvent) -> None:
        self.enqueue("_save_attribute", event)

    async def _save_attribute(
        self,
        event: AttributeReadEvent
        | AttributeReportedEvent
        | AttributeUpdatedEvent
        | AttributeWrittenEvent,
    ) -> None:
        if isinstance(event, AttributeWrittenEvent) and event.status != Status.SUCCESS:
            LOGGER.debug("Ignoring failed attribute write event: %s", event)
            return

        await self.execute(
            f"""
            INSERT INTO attributes_cache{DB_V} (ieee, endpoint_id, cluster_type, cluster_id, attr_id, manufacturer_code, status, value, last_updated)
            VALUES (:ieee, :endpoint_id, :cluster_type, :cluster_id, :attr_id, :manufacturer_code, :status, :value, :timestamp)
                ON CONFLICT (ieee, endpoint_id, cluster_type, cluster_id, attr_id, manufacturer_code_idx) DO UPDATE
                SET status=excluded.status, value=excluded.value, last_updated=excluded.last_updated
                WHERE
                    value != excluded.value
                    OR status != excluded.status
                    OR :timestamp - last_updated > :min_update_delta
            """,
            {
                "ieee": event.device_ieee,
                "endpoint_id": event.endpoint_id,
                "cluster_type": event.cluster_type,
                "cluster_id": event.cluster_id,
                "attr_id": event.attribute_id,
                "manufacturer_code": event.manufacturer_code,
                "status": Status.SUCCESS,
                "value": event.value,
                "timestamp": datetime.now(UTC).timestamp(),
                "min_update_delta": MIN_UPDATE_DELTA,
            },
        )

        await self._db.commit()

    def on_attribute_cleared(self, event: AttributeClearedEvent) -> None:
        self.enqueue("_clear_attribute", event)

    async def _clear_attribute(self, event: AttributeClearedEvent) -> None:
        q = f"""
            DELETE FROM attributes_cache{DB_V}
            WHERE
                ieee = :ieee
                AND endpoint_id = :endpoint_id
                AND cluster_type = :cluster_type
                AND cluster_id = :cluster_id
                AND attr_id = :attr_id
                AND manufacturer_code IS NOT DISTINCT FROM :manufacturer_code
            """

        await self.execute(
            q,
            {
                "ieee": event.device_ieee,
                "endpoint_id": event.endpoint_id,
                "cluster_type": event.cluster_type,
                "cluster_id": event.cluster_id,
                "attr_id": event.attribute_id,
                "manufacturer_code": event.manufacturer_code,
            },
        )
        await self._db.commit()

    def on_attribute_unsupported(self, event: AttributeUnsupportedEvent) -> None:
        self.enqueue("_unsupported_attribute_added", event)

    async def _unsupported_attribute_added(
        self, event: AttributeUnsupportedEvent
    ) -> None:
        q = f"""INSERT INTO attributes_cache{DB_V} (ieee, endpoint_id, cluster_type, cluster_id, attr_id, manufacturer_code, status, value, last_updated)
                   VALUES (:ieee, :endpoint_id, :cluster_type, :cluster_id, :attr_id, :manufacturer_code, :status, :value, :timestamp)
                   ON CONFLICT (ieee, endpoint_id, cluster_type, cluster_id, attr_id, manufacturer_code_idx)
                   DO UPDATE SET status=excluded.status, value=excluded.value, last_updated=excluded.last_updated"""

        await self.execute(
            q,
            {
                "ieee": event.device_ieee,
                "endpoint_id": event.endpoint_id,
                "cluster_type": event.cluster_type,
                "cluster_id": event.cluster_id,
                "attr_id": event.attribute_id,
                "manufacturer_code": event.manufacturer_code,
                "status": Status.UNSUPPORTED_ATTRIBUTE,
                "value": None,
                "timestamp": datetime.now(UTC).timestamp(),
            },
        )
        await self._db.commit()

    def on_ota_query_cache_updated(self, event: OtaQueryCacheUpdatedEvent) -> None:
        self.enqueue("_save_ota_query_cache_entry", event)

    async def _save_ota_query_cache_entry(
        self, event: OtaQueryCacheUpdatedEvent
    ) -> None:
        q = f"""INSERT INTO ota_query_cache{DB_V}
                    (ieee, endpoint_id, manufacturer_code, image_type,
                     current_file_version, hardware_version, last_updated)
                VALUES (?, ?, ?, ?, ?, ?, ?)
                ON CONFLICT (ieee, endpoint_id) DO UPDATE SET
                    manufacturer_code=excluded.manufacturer_code,
                    image_type=excluded.image_type,
                    current_file_version=excluded.current_file_version,
                    hardware_version=excluded.hardware_version,
                    last_updated=excluded.last_updated"""

        await self.execute(
            q,
            (
                event.device_ieee,
                event.endpoint_id,
                event.manufacturer_code,
                event.image_type,
                event.current_file_version,
                event.hardware_version,
                datetime.now(UTC).timestamp(),
            ),
        )
        await self._db.commit()

    def on_ota_query_cache_cleared(self, event: OtaQueryCacheClearedEvent) -> None:
        self.enqueue("_delete_ota_query_cache_entry", event)

    async def _delete_ota_query_cache_entry(
        self, event: OtaQueryCacheClearedEvent
    ) -> None:
        await self.execute(
            f"DELETE FROM ota_query_cache{DB_V} WHERE ieee = ? AND endpoint_id = ?",
            (event.device_ieee, event.endpoint_id),
        )
        await self._db.commit()

    def network_backup_created(self, backup: zigpy.backups.NetworkBackup) -> None:
        self.enqueue("_network_backup_created", json.dumps(backup.as_dict()))

    async def _network_backup_created(self, backup_json: str) -> None:
        q = f"""INSERT INTO network_backups{DB_V} VALUES (?, ?)
                    ON CONFLICT (id)
                    DO UPDATE SET
                        backup_json=excluded.backup_json"""

        await self.execute(q, (None, backup_json))
        await self._db.commit()

    def network_backup_removed(self, backup: zigpy.backups.NetworkBackup) -> None:
        self.enqueue("_network_backup_removed", backup.backup_time)

    async def _network_backup_removed(self, backup_time: datetime) -> None:
        q = f"""DELETE FROM network_backups{DB_V}
                    WHERE json_extract(backup_json, '$.backup_time')=?"""

        await self.execute(q, (backup_time.isoformat(),))
        await self._db.commit()

    async def _read_all_attributes(
        self,
    ) -> list[AttributeCacheRow]:
        """Read all attribute rows from the database."""
        async with self.execute(
            f"""
            SELECT ieee, endpoint_id, cluster_type, cluster_id, attr_id,
                   manufacturer_code, status, value, last_updated
            FROM attributes_cache{DB_V}
            """
        ) as cursor:
            return [AttributeCacheRow(*row) for row in await cursor.fetchall()]

    async def load(self) -> None:
        LOGGER.debug("Loading application state")
        await self._load_devices()
        await self._load_node_descriptors()
        await self._load_endpoints()
        await self._load_clusters()

        # Read all attribute rows from the database once
        all_attributes = await self._read_all_attributes()

        # First pass: populate cache on bare clusters for quirks
        await self._populate_attribute_cache(all_attributes)

        for device in self._application.devices.values():
            # Populate the device signature before we apply any quirks, which can modify
            # the device structure (for now)
            device.original_signature = device.get_signature()
            new_device = self._application._resolve_device(device)
            new_device.original_signature = device.original_signature
            self._application.devices[device.ieee] = new_device

        # Clear the attribute cache to ensure the quirked state is correct
        for device in self._application.devices.values():
            for ep in device.non_zdo_endpoints:
                for cluster in ep.in_clusters.values():
                    cluster._attr_cache.clear()

                for cluster in ep.out_clusters.values():
                    cluster._attr_cache.clear()

        # Second pass: populate the attribute cache for the final device state and
        # migrate attributes with unknown manufacturer codes to the correct codes. Only
        # the migration pass modifies the database so we do it in a transaction.
        async with self._transaction():
            await self._populate_attribute_cache(all_attributes, migrate=True)

        await self._load_groups()
        await self._load_group_members()
        await self._load_relays()
        await self._load_neighbors()
        await self._load_routes()
        await self._load_network_backups()
        await self._load_ota_query_cache()

        await self._db.commit()

        await self._register_device_listeners()

    async def _populate_attribute_cache(
        self,
        rows: list[AttributeCacheRow],
        *,
        migrate: bool = False,
    ) -> None:
        """Populate cluster attribute cache from pre-loaded rows.

        When `migrate` is True, unmigrated rows with ambiguous attribute IDs are
        resolved using the (now-quirked) cluster definitions and the database is
        updated to match.
        """
        for row in rows:
            dev = self._application.get_device(row.ieee)

            LOGGER.debug(
                "[0x%04x:%s:0x%04x] Loading attribute %s=%r status=%r mfg_code=%r",
                dev.nwk,
                row.endpoint_id,
                row.cluster_id,
                (
                    row.attr_id
                    if isinstance(row.attr_id, str)
                    else f"0x{row.attr_id:04x}"
                ),
                row.value,
                row.status,
                row.manufacturer_code,
            )

            if row.endpoint_id not in dev.endpoints:
                continue

            ep = dev.endpoints[row.endpoint_id]
            clusters = (
                ep.in_clusters
                if row.cluster_type == ClusterType.Server
                else ep.out_clusters
            )

            if row.cluster_id not in clusters:
                LOGGER.debug("Unknown ZCL cluster, skipping")
                continue

            cluster = clusters[row.cluster_id]

            # For unmigrated rows on the second pass, try to resolve the
            # manufacturer code using the full quirk cluster definitions
            manufacturer_code = row.manufacturer_code

            if migrate and row.manufacturer_code == UNMIGRATED_MANUFACTURER_CODE:
                resolved_manufacturer_code = self._resolve_unmigrated_attribute(
                    cluster=cluster,
                    attr_id=row.attr_id,
                    value=row.value,
                    dev=dev,
                )

                if resolved_manufacturer_code is not UNDEFINED:
                    manufacturer_code = resolved_manufacturer_code

                    row_params = {
                        "ieee": row.ieee,
                        "endpoint_id": row.endpoint_id,
                        "cluster_type": row.cluster_type,
                        "cluster_id": row.cluster_id,
                        "attr_id": row.attr_id,
                    }

                    # Delete the unmigrated row and re-insert with the resolved
                    # manufacturer code. INSERT OR IGNORE handles the case where a row
                    # with the resolved code already exists (e.g. the user manually
                    # read the attribute through the UI).
                    await self.execute(
                        f"""
                        DELETE FROM attributes_cache{DB_V}
                        WHERE
                            ieee = :ieee
                            AND endpoint_id = :endpoint_id
                            AND cluster_type = :cluster_type
                            AND cluster_id = :cluster_id
                            AND attr_id = :attr_id
                            AND manufacturer_code = :old_manufacturer_code
                        """,
                        {
                            **row_params,
                            "old_manufacturer_code": UNMIGRATED_MANUFACTURER_CODE,
                        },
                    )

                    async with self.execute(
                        f"""
                        INSERT OR IGNORE INTO attributes_cache{DB_V}
                            (ieee, endpoint_id, cluster_type, cluster_id,
                             attr_id, manufacturer_code, status, value,
                             last_updated)
                        VALUES
                            (:ieee, :endpoint_id, :cluster_type, :cluster_id,
                             :attr_id, :manufacturer_code, :status, :value,
                             :last_updated)
                        """,
                        {
                            **row_params,
                            "manufacturer_code": manufacturer_code,
                            "status": row.status,
                            "value": row.value,
                            "last_updated": row.last_updated,
                        },
                    ) as cursor:
                        # A resolved row already exists, it will populate the cache
                        if cursor.rowcount == 0:
                            continue

            try:
                attr_def = cluster.find_attribute(
                    row.attr_id,
                    manufacturer_code=(
                        UNDEFINED
                        if manufacturer_code == UNMIGRATED_MANUFACTURER_CODE
                        else manufacturer_code
                    ),
                )
            except KeyError:
                LOGGER.debug("Unknown ZCL attribute, skipping")

                # Unsupported unknown attributes are dropped
                if row.status == Status.SUCCESS:
                    cluster._attr_cache.set_legacy_value(
                        row.attr_id,
                        row.value,
                        last_updated=datetime.fromtimestamp(row.last_updated, UTC),
                    )

                continue

            if row.status == Status.SUCCESS:
                cluster._attr_cache.set_value(
                    attr_def,
                    row.value,
                    last_updated=datetime.fromtimestamp(row.last_updated, UTC),
                )
            else:
                cluster._attr_cache.mark_unsupported(attr_def)
                continue

            # Populate the device's manufacturer and model attributes
            if (
                row.cluster_id == Basic.cluster_id
                and attr_def == Basic.AttributeDefs.manufacturer
            ):
                dev.manufacturer = decode_str_attribute(row.value)
            elif (
                row.cluster_id == Basic.cluster_id
                and attr_def == Basic.AttributeDefs.model
            ):
                dev.model = decode_str_attribute(row.value)

    @staticmethod
    def _resolve_unmigrated_attribute(
        cluster: Cluster,
        attr_id: int,
        value: Any,
        dev: Device,
    ) -> int | None | zigpy.typing.UndefinedType:
        """Try to resolve the manufacturer code for an unmigrated attribute.

        When multiple attribute definitions share an ID, the manufacturer-specific one
        is preferred if there are exactly two candidates (one standard, one
        manufacturer-specific). Otherwise the attribute is considered ambiguous and
        skipped (left unmigrated).

        Returns the resolved manufacturer code (including `None`), or UNDEFINED if
        unresolvable.
        """
        try:
            attr_defs = cluster.find_attributes(attr_id)
        except KeyError:
            LOGGER.debug(
                "Unable to find any attributes %r=%r on cluster %r for %r"
                " for data migration, skipping",
                attr_id,
                value,
                cluster,
                dev,
            )
            return UNDEFINED

        manuf_attrs = [
            a
            for a in attr_defs
            if cluster._get_effective_manufacturer_code(a) is not None
        ]

        if len(attr_defs) == 1:
            attr_def = attr_defs[0]
        elif len(attr_defs) == 2 and len(manuf_attrs) == 1:
            # One standard + one manufacturer-specific: prefer manufacturer-specific
            attr_def = manuf_attrs[0]
        else:
            LOGGER.debug(
                "Unable to find unique attribute %r=%r on cluster %r for %r"
                " for data migration, skipping (candidates: %r)",
                attr_id,
                value,
                cluster,
                dev,
                attr_defs,
            )
            return UNDEFINED

        return cluster._get_effective_manufacturer_code(attr_def)

    async def _load_devices(self) -> None:
        async with self.execute(f"SELECT * FROM devices{DB_V}") as cursor:
            async for ieee, nwk, status, last_seen in cursor:
                dev = self._application.add_device(ieee, nwk)
                dev.status = DeviceStatus(status)

                if last_seen > 0:
                    dev.last_seen = last_seen

    async def _load_node_descriptors(self) -> None:
        async with self.execute(f"SELECT * FROM node_descriptors{DB_V}") as cursor:
            async for ieee, *fields in cursor:
                dev = self._application.get_device(ieee)
                dev.node_desc = zdo_t.NodeDescriptor(*fields)
                assert dev.node_desc.is_valid

    async def _load_endpoints(self) -> None:
        async with self.execute(f"SELECT * FROM endpoints{DB_V}") as cursor:
            async for ieee, epid, profile_id, device_type, status in cursor:
                dev = self._application.get_device(ieee)
                ep = dev.add_endpoint(epid)
                ep.profile_id = profile_id
                ep.status = EndpointStatus(status)

                if profile_id == zigpy.profiles.zha.PROFILE_ID:
                    ep.device_type = zigpy.profiles.zha.DeviceType(device_type)
                elif profile_id == zigpy.profiles.zll.PROFILE_ID:
                    ep.device_type = zigpy.profiles.zll.DeviceType(device_type)
                else:
                    ep.device_type = device_type

    async def _load_clusters(self) -> None:
        async with self.execute(f"SELECT * FROM clusters{DB_V}") as cursor:
            async for ieee, endpoint_id, cluster_type, cluster_id in cursor:
                dev = self._application.get_device(ieee)
                ep = dev.endpoints[endpoint_id]

                if ClusterType(cluster_type) == ClusterType.Server:
                    ep.add_input_cluster(cluster_id)
                else:
                    ep.add_output_cluster(cluster_id)

    async def _load_groups(self) -> None:
        async with self.execute(f"SELECT * FROM groups{DB_V}") as cursor:
            async for group_id, name in cursor:
                self._application.groups.add_group(group_id, name, suppress_event=True)

    async def _load_group_members(self) -> None:
        async with self.execute(f"SELECT * FROM group_members{DB_V}") as cursor:
            async for group_id, ieee, ep_id in cursor:
                dev = self._application.get_device(ieee)
                group = self._application.groups[group_id]
                group.add_member(dev.endpoints[ep_id], suppress_event=True)

    async def _load_relays(self) -> None:
        async with self.execute(f"SELECT * FROM relays{DB_V}") as cursor:
            async for ieee, value in cursor:
                dev = self._application.get_device(ieee)
                relays, _ = t.Relays.deserialize(value)
                dev.relays = zigpy.util.filter_relays(relays)

    async def _load_neighbors(self) -> None:
        async with self.execute(f"SELECT * FROM neighbors{DB_V}") as cursor:
            async for ieee, *fields in cursor:
                neighbor = zdo_t.Neighbor(*fields)
                self._application.topology.neighbors[ieee].append(neighbor)

    async def _load_routes(self) -> None:
        async with self.execute(f"SELECT * FROM routes{DB_V}") as cursor:
            async for ieee, *fields in cursor:
                route = zdo_t.Route(*fields)
                self._application.topology.routes[ieee].append(route)

    async def _load_network_backups(self) -> None:
        self._application.backups.backups.clear()

        async with self.execute(
            f"SELECT * FROM network_backups{DB_V} ORDER BY id"
        ) as cursor:
            backups = []

            async for _id, backup_json in cursor:
                backup = zigpy.backups.NetworkBackup.from_dict(json.loads(backup_json))
                backups.append(backup)

        backups.sort(key=lambda b: b.backup_time)

        for backup in backups:
            self._application.backups.add_backup(backup, suppress_event=True)

    async def _load_ota_query_cache(self) -> None:
        async with self.execute(f"SELECT * FROM ota_query_cache{DB_V}") as cursor:
            async for (
                ieee,
                endpoint_id,
                manufacturer_code,
                image_type,
                current_file_version,
                hardware_version,
                _last_updated,
            ) in cursor:
                try:
                    dev = self._application.get_device(ieee)
                    ep = dev.endpoints[endpoint_id]
                except KeyError:
                    # Quirks or firmware updates can remove endpoints/clusters
                    continue

                field_control = (
                    Ota.QueryNextImageCommand.FieldControl(0)
                    if hardware_version is None
                    else Ota.QueryNextImageCommand.FieldControl.HardwareVersion
                )
                cmd = Ota.QueryNextImageCommand(
                    field_control=field_control,
                    manufacturer_code=manufacturer_code,
                    image_type=image_type,
                    current_file_version=current_file_version,
                )
                if hardware_version is not None:
                    cmd.hardware_version = hardware_version

                # Restore to the first OTA cluster found; prefer client
                # (out_clusters) since that's where runtime routing places it
                # when both cluster types exist.
                for clusters in (ep.out_clusters, ep.in_clusters):
                    cluster = clusters.get(Ota.cluster_id)
                    if isinstance(cluster, Ota):
                        cluster.last_query_cmd = cmd
                        break

    async def _register_device_listeners(self) -> None:
        for dev in self._application.devices.values():
            dev.add_context_listener(self)

    @contextlib.asynccontextmanager
    async def _transaction(self):
        await self.execute("BEGIN TRANSACTION")

        try:
            yield
        except Exception:  # noqa: BLE001
            await self.execute("ROLLBACK")
            raise
        else:
            await self.execute("COMMIT")

    async def _get_table_versions(self) -> dict[str, int]:
        tables = {}

        async with self.execute(
            "SELECT name FROM sqlite_master WHERE type='table'"
        ) as cursor:
            async for (name,) in cursor:
                # Ignore tables internal to SQLite
                if name.startswith("sqlite_"):
                    continue

                # The regex will always return a match
                match = DB_V_REGEX.search(name)
                assert match is not None

                tables[name] = int(match.group(0)[2:] or "0")

        return tables

    async def _table_exists(self, name: str) -> bool:
        return name in (await self._get_table_versions())

    async def _run_migrations(self) -> bool:
        """Migrates the database to the newest schema, returning True if migrations ran."""

        tables = await self._get_table_versions()
        tables_version = max(tables.values(), default=0)

        async with self.execute("PRAGMA user_version") as cursor:
            (db_version,) = await cursor.fetchone()

        LOGGER.debug(
            "Current database version is v%s (table version v%s)",
            db_version,
            tables_version,
        )

        # Table version suffixes were introduced in v4. If the table version suffix does
        # not match `user_version`, either zigpy was downgraded to a *really* old
        # version (July 2021), or it's corrupt. Running migrations could delete existing
        # table data, and since we cannot guarantee the schema is intact, fail early.
        if tables_version >= 4 and tables_version != db_version:
            raise zigpy.exceptions.CorruptDatabase(
                f"The `zigbee.db` database version ({db_version}) does not match its"
                f" max table version ({tables_version}). The database is inconsistent.",
            )

        if db_version == 0 and not tables:
            # If this is a brand new database, just load the current schema
            await self.executescript(zigpy.appdb_schemas.SCHEMAS[DB_VERSION])
            return False
        elif db_version > DB_VERSION:
            LOGGER.error(
                "This zigpy release uses database schema v%s but the database is v%s."
                " Downgrading zigpy is *not* recommended and may result in data loss."
                " Use at your own risk.",
                DB_VERSION,
                db_version,
            )
            return False

        # All migrations must succeed. If any fail, the database is not touched.
        async with self._transaction():
            for migration, to_db_version in [
                (self._migrate_to_v4, 4),
                (self._migrate_to_v5, 5),
                (self._migrate_to_v6, 6),
                (self._migrate_to_v7, 7),
                (self._migrate_to_v8, 8),
                (self._migrate_to_v9, 9),
                (self._migrate_to_v10, 10),
                (self._migrate_to_v11, 11),
                (self._migrate_to_v12, 12),
                (self._migrate_to_v13, 13),
                (self._migrate_to_v14, 14),
                (self._migrate_to_v15, 15),
            ]:
                if db_version >= min(to_db_version, DB_VERSION):
                    continue

                LOGGER.info(
                    "Migrating database from v%d to v%d", db_version, to_db_version
                )
                await self.executescript(zigpy.appdb_schemas.SCHEMAS[to_db_version])
                await migration()

                db_version = to_db_version

        return True

    async def _migrate_tables(
        self, table_map: dict[str, str | None], *, errors: str = "raise"
    ):
        """Copy rows from one set of tables into another."""

        # Extract the "old" table version suffix
        tables = await self._get_table_versions()
        old_table_name = list(table_map.keys())[0]
        old_version = tables[old_table_name]

        # Check which tables would not be migrated
        old_tables = [t for t, v in tables.items() if v == old_version]
        unmigrated_old_tables = [t for t in old_tables if t not in table_map]

        if unmigrated_old_tables:
            raise RuntimeError(
                f"The following tables were not migrated: {unmigrated_old_tables}"
            )

        # Insertion order matters for foreign key constraints but any rows that fail
        # to insert due to constraint violations can be discarded
        for old_table, new_table in table_map.items():
            # Ignore tables without a migration
            if new_table is None:
                continue

            # Use explicit column names to skip generated columns automatically
            async with self.execute(f"PRAGMA table_info({old_table})") as cursor:
                columns = [row[1] async for row in cursor]

            col_list = ", ".join(columns)
            placeholders = ", ".join("?" * len(columns))
            select_sql = f"SELECT {col_list} FROM {old_table}"
            insert_sql = f"INSERT INTO {new_table} ({col_list}) VALUES ({placeholders})"

            async with self.execute(select_sql) as cursor:
                async for row in cursor:
                    try:
                        await self.execute(insert_sql, row)
                    except sqlite3.IntegrityError as e:
                        if errors == "raise":
                            raise
                        elif errors == "warn":
                            LOGGER.warning(
                                "Failed to migrate row %s%s: %s", old_table, row, e
                            )
                        elif errors == "ignore":
                            pass
                        else:
                            raise ValueError(
                                f"Invalid value for `errors`: {errors!r}"
                            ) from e

    async def _migrate_to_v4(self):
        """Schema v4 expanded the node descriptor and neighbor table columns"""
        # The `node_descriptors` table was added in v1
        if await self._table_exists("node_descriptors"):
            async with self.execute("SELECT * FROM node_descriptors") as cur:
                async for dev_ieee, value in cur:
                    node_desc, rest = zdo_t.NodeDescriptor.deserialize(value)
                    assert not rest

                    await self.execute(
                        "INSERT INTO node_descriptors_v4"
                        " VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?)",
                        (dev_ieee, *node_desc.as_tuple()),
                    )

        # The `neighbors` table was added in v3 but the version number was not
        # incremented. It may not exist.
        if await self._table_exists("neighbors"):
            async with self.execute("SELECT * FROM neighbors") as cur:
                async for dev_ieee, epid, ieee, nwk, packed, prm, depth, lqi in cur:
                    neighbor = zdo_t.Neighbor(
                        extended_pan_id=epid,
                        ieee=ieee,
                        nwk=nwk,
                        permit_joining=prm,
                        depth=depth,
                        lqi=lqi,
                        reserved2=0b000000,
                        **zdo_t.Neighbor._parse_packed(packed),
                    )

                    await self.execute(
                        "INSERT INTO neighbors_v4 VALUES (?,?,?,?,?,?,?,?,?,?,?,?)",
                        (dev_ieee, *neighbor.as_tuple()),
                    )

    async def _migrate_to_v5(self):
        """Schema v5 introduced global table version suffixes and removed stale rows"""

        await self._migrate_tables(
            {
                "devices": "devices_v5",
                "endpoints": "endpoints_v5",
                "clusters": "in_clusters_v5",
                "output_clusters": "out_clusters_v5",
                "groups": "groups_v5",
                "group_members": "group_members_v5",
                "relays": "relays_v5",
                "attributes": "attributes_cache_v5",
                # These were migrated in v4
                "neighbors_v4": "neighbors_v5",
                "node_descriptors_v4": "node_descriptors_v5",
                # Explicitly specify which tables will not be migrated
                "neighbors": None,
                "node_descriptors": None,
            },
            errors="warn",
        )

    async def _migrate_to_v6(self):
        """Schema v6 relaxed the `attribute_cache` table schema to ignore endpoints"""

        await self._migrate_tables(
            {
                "devices_v5": "devices_v6",
                "endpoints_v5": "endpoints_v6",
                "in_clusters_v5": "in_clusters_v6",
                "out_clusters_v5": "out_clusters_v6",
                "groups_v5": "groups_v6",
                "group_members_v5": "group_members_v6",
                "relays_v5": "relays_v6",
                "attributes_cache_v5": "attributes_cache_v6",
                "neighbors_v5": "neighbors_v6",
                "node_descriptors_v5": "node_descriptors_v6",
            }
        )

        # See if we can migrate any `attributes_cache` rows skipped by the v5 migration
        if await self._table_exists("attributes"):
            async with self.execute("SELECT count(*) FROM attributes") as cur:
                (num_attrs_v4,) = await cur.fetchone()

            async with self.execute("SELECT count(*) FROM attributes_cache_v6") as cur:
                (num_attrs_v6,) = await cur.fetchone()

            if num_attrs_v6 < num_attrs_v4:
                LOGGER.warning(
                    "Migrating up to %d rows skipped by v5 migration",
                    num_attrs_v4 - num_attrs_v6,
                )

                await self._migrate_tables(
                    {
                        "attributes": "attributes_cache_v6",
                        "devices": None,
                        "endpoints": None,
                        "clusters": None,
                        "neighbors": None,
                        "node_descriptors": None,
                        "output_clusters": None,
                        "groups": None,
                        "group_members": None,
                        "relays": None,
                    },
                    errors="ignore",
                )

    async def _migrate_to_v7(self):
        """Schema v7 added the `unsupported_attributes` table."""

        await self._migrate_tables(
            {
                "devices_v6": "devices_v7",
                "endpoints_v6": "endpoints_v7",
                "in_clusters_v6": "in_clusters_v7",
                "out_clusters_v6": "out_clusters_v7",
                "groups_v6": "groups_v7",
                "group_members_v6": "group_members_v7",
                "relays_v6": "relays_v7",
                "attributes_cache_v6": "attributes_cache_v7",
                "neighbors_v6": "neighbors_v7",
                "node_descriptors_v6": "node_descriptors_v7",
            }
        )

    async def _migrate_to_v8(self):
        """Schema v8 added the `devices_v8.last_seen` column."""

        async with self.execute("SELECT * FROM devices_v7") as cursor:
            async for ieee, nwk, status in cursor:
                # Set the default `last_seen` to the unix epoch
                await self.execute(
                    "INSERT INTO devices_v8 VALUES (?, ?, ?, ?)",
                    (ieee, nwk, status, 0),
                )

        # Copy the devices table first, it should have no conflicts
        await self._migrate_tables(
            {
                "endpoints_v7": "endpoints_v8",
                "in_clusters_v7": "in_clusters_v8",
                "out_clusters_v7": "out_clusters_v8",
                "groups_v7": "groups_v8",
                "group_members_v7": "group_members_v8",
                "relays_v7": "relays_v8",
                "attributes_cache_v7": "attributes_cache_v8",
                "neighbors_v7": "neighbors_v8",
                "node_descriptors_v7": "node_descriptors_v8",
                "unsupported_attributes_v7": "unsupported_attributes_v8",
                "devices_v7": None,
            }
        )

    async def _migrate_to_v9(self):
        """Schema v9 changed the data type of the `devices_v8.last_seen` column."""

        await self.execute(
            """INSERT INTO devices_v9 (ieee, nwk, status, last_seen)
            SELECT ieee, nwk, status, last_seen / 1000.0 FROM devices_v8"""
        )

        await self._migrate_tables(
            {
                "endpoints_v8": "endpoints_v9",
                "in_clusters_v8": "in_clusters_v9",
                "out_clusters_v8": "out_clusters_v9",
                "groups_v8": "groups_v9",
                "group_members_v8": "group_members_v9",
                "relays_v8": "relays_v9",
                "attributes_cache_v8": "attributes_cache_v9",
                "neighbors_v8": "neighbors_v9",
                "node_descriptors_v8": "node_descriptors_v9",
                "unsupported_attributes_v8": "unsupported_attributes_v9",
                "devices_v8": None,
            }
        )

    async def _migrate_to_v10(self):
        """Schema v10 added a new `network_backups_v10` table."""

        await self._migrate_tables(
            {
                "devices_v9": "devices_v10",
                "endpoints_v9": "endpoints_v10",
                "in_clusters_v9": "in_clusters_v10",
                "out_clusters_v9": "out_clusters_v10",
                "groups_v9": "groups_v10",
                "group_members_v9": "group_members_v10",
                "relays_v9": "relays_v10",
                "attributes_cache_v9": "attributes_cache_v10",
                "neighbors_v9": "neighbors_v10",
                "node_descriptors_v9": "node_descriptors_v10",
                "unsupported_attributes_v9": "unsupported_attributes_v10",
            }
        )

    async def _migrate_to_v11(self):
        """Schema v11 added a new `routes_v11` table."""

        await self._migrate_tables(
            {
                "devices_v10": "devices_v11",
                "endpoints_v10": "endpoints_v11",
                "in_clusters_v10": "in_clusters_v11",
                "out_clusters_v10": "out_clusters_v11",
                "groups_v10": "groups_v11",
                "group_members_v10": "group_members_v11",
                "relays_v10": "relays_v11",
                "attributes_cache_v10": "attributes_cache_v11",
                "neighbors_v10": "neighbors_v11",
                "node_descriptors_v10": "node_descriptors_v11",
                "unsupported_attributes_v10": "unsupported_attributes_v11",
                "network_backups_v10": "network_backups_v11",
            }
        )

    async def _migrate_to_v12(self):
        """Schema v12 added a `timestamp` column to attribute updates."""

        await self._migrate_tables(
            {
                "devices_v11": "devices_v12",
                "endpoints_v11": "endpoints_v12",
                "in_clusters_v11": "in_clusters_v12",
                "neighbors_v11": "neighbors_v12",
                "routes_v11": "routes_v12",
                "node_descriptors_v11": "node_descriptors_v12",
                "out_clusters_v11": "out_clusters_v12",
                "groups_v11": "groups_v12",
                "group_members_v11": "group_members_v12",
                "relays_v11": "relays_v12",
                "unsupported_attributes_v11": "unsupported_attributes_v12",
                "network_backups_v11": "network_backups_v12",
                "attributes_cache_v11": None,
            }
        )

        async with self.execute("SELECT * FROM attributes_cache_v11") as cursor:
            async for ieee, endpoint_id, cluster_id, attrid, value in cursor:
                # Set the default `last_updated` to the unix epoch
                await self.execute(
                    "INSERT INTO attributes_cache_v12 VALUES (?, ?, ?, ?, ?, ?)",
                    (ieee, endpoint_id, cluster_id, attrid, value, 0),
                )

    async def _migrate_to_v13(self):
        """Schema v13 combines both cluster types and caching for all attributes."""

        await self._migrate_tables(
            {
                "devices_v12": "devices_v13",
                "endpoints_v12": "endpoints_v13",
                "neighbors_v12": "neighbors_v13",
                "routes_v12": "routes_v13",
                "node_descriptors_v12": "node_descriptors_v13",
                "groups_v12": "groups_v13",
                "group_members_v12": "group_members_v13",
                "relays_v12": "relays_v13",
                "network_backups_v12": "network_backups_v13",
                "in_clusters_v12": None,
                "out_clusters_v12": None,
                "unsupported_attributes_v12": None,
                "attributes_cache_v12": None,
            }
        )

        async with self.execute("SELECT * FROM in_clusters_v12") as cursor:
            async for ieee, endpoint_id, cluster_id in cursor:
                await self.execute(
                    "INSERT INTO clusters_v13 VALUES (?, ?, ?, ?)",
                    (ieee, endpoint_id, ClusterType.Server, cluster_id),
                )

        async with self.execute("SELECT * FROM out_clusters_v12") as cursor:
            async for ieee, endpoint_id, cluster_id in cursor:
                await self.execute(
                    "INSERT INTO clusters_v13 VALUES (?, ?, ?, ?)",
                    (ieee, endpoint_id, ClusterType.Client, cluster_id),
                )

        async with self.execute("SELECT * FROM unsupported_attributes_v12") as cursor:
            async for ieee, endpoint_id, cluster_id, attrid in cursor:
                await self.execute(
                    "INSERT INTO unsupported_attributes_v13 VALUES (?, ?, ?, ?, ?)",
                    (ieee, endpoint_id, ClusterType.Server, cluster_id, attrid),
                )

        async with self.execute("SELECT * FROM attributes_cache_v12") as cursor:
            async for (
                ieee,
                endpoint_id,
                cluster_id,
                attrid,
                value,
                last_updated,
            ) in cursor:
                await self.execute(
                    "INSERT INTO attributes_cache_v13 VALUES (?, ?, ?, ?, ?, ?, ?)",
                    (
                        ieee,
                        endpoint_id,
                        ClusterType.Server,
                        cluster_id,
                        attrid,
                        value,
                        last_updated,
                    ),
                )

    async def _migrate_to_v14(self) -> None:
        """Schema v14 adds `manufacturer_code` and `status` to the attribute cache."""
        await self._migrate_tables(
            {
                "devices_v13": "devices_v14",
                "endpoints_v13": "endpoints_v14",
                "neighbors_v13": "neighbors_v14",
                "routes_v13": "routes_v14",
                "node_descriptors_v13": "node_descriptors_v14",
                "groups_v13": "groups_v14",
                "group_members_v13": "group_members_v14",
                "relays_v13": "relays_v14",
                "network_backups_v13": "network_backups_v14",
                "clusters_v13": "clusters_v14",
                "unsupported_attributes_v13": None,
                "attributes_cache_v13": None,
            }
        )

        # Migrate unsupported attributes into the attributes cache with status
        async with self.execute("SELECT * FROM unsupported_attributes_v13") as cursor:
            async for (
                ieee,
                endpoint_id,
                cluster_type,
                cluster_id,
                attrid,
            ) in cursor:
                await self.execute(
                    "INSERT INTO attributes_cache_v14 (ieee, endpoint_id, cluster_type, cluster_id, attr_id, manufacturer_code, status, value, last_updated) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
                    (
                        ieee,
                        endpoint_id,
                        cluster_type,
                        cluster_id,
                        attrid,
                        UNMIGRATED_MANUFACTURER_CODE,
                        Status.UNSUPPORTED_ATTRIBUTE,
                        None,
                        datetime.fromtimestamp(0, UTC).timestamp(),
                    ),
                )

        async with self.execute("SELECT * FROM attributes_cache_v13") as cursor:
            async for (
                ieee,
                endpoint_id,
                cluster_type,
                cluster_id,
                attrid,
                value,
                last_updated,
            ) in cursor:
                # Use INSERT OR IGNORE because the same attribute may exist in both
                # unsupported_attributes_v13 and attributes_cache_v13. The unsupported
                # status (inserted first) should win over old cached values.
                await self.execute(
                    "INSERT OR IGNORE INTO attributes_cache_v14 (ieee, endpoint_id, cluster_type, cluster_id, attr_id, manufacturer_code, status, value, last_updated) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
                    (
                        ieee,
                        endpoint_id,
                        cluster_type,
                        cluster_id,
                        attrid,
                        UNMIGRATED_MANUFACTURER_CODE,
                        Status.SUCCESS,
                        value,
                        last_updated,
                    ),
                )

    async def _migrate_to_v15(self) -> None:
        """Schema v15 adds `ota_query_cache` table for persisting OTA query fields."""
        await self._migrate_tables(
            {
                "devices_v14": "devices_v15",
                "endpoints_v14": "endpoints_v15",
                "neighbors_v14": "neighbors_v15",
                "routes_v14": "routes_v15",
                "node_descriptors_v14": "node_descriptors_v15",
                "groups_v14": "groups_v15",
                "group_members_v14": "group_members_v15",
                "relays_v14": "relays_v15",
                "network_backups_v14": "network_backups_v15",
                "clusters_v14": "clusters_v15",
                "attributes_cache_v14": "attributes_cache_v15",
            }
        )
        # ota_query_cache_v15 is new and starts empty
